import re
import tkinter as tk
import tkinter.ttk as ttk
from abc import abstractmethod
from decimal import Decimal

import numpy as np
from numpy import ma

from . import moduleFrame
from .scrolledFrame import ScrolledFrame
from .table import ButtonFrame, Table, WrappedLabel

DEFAULT_INITIAL_CONC = 1e-4

prefixesDecimal = {
    "": Decimal(1),
    "m": Decimal(1e-3),
    "u": Decimal(1e-6),
    "μ": Decimal(1e-6),
    "n": Decimal(1e-9),
}

prefixes = dict([key, float(value)] for key, value in prefixesDecimal.items())


def convertConc(conc, fromUnit, toUnit):
    if conc is ma.masked:
        return ""
    conc = Decimal(conc)
    convertedConc = float(
        conc * prefixesDecimal[fromUnit.strip("M")] / prefixesDecimal[toUnit.strip("M")]
    )
    return f"{convertedConc:g}"  # strip trailing zeroes


class totalConcentrations(moduleFrame.Strategy):
    requiredAttributes = (
        "concsUnit",
        "totalConcs",
        "freeNames",
        "variableNames",
    )

    # TODO: support entering initial guesses
    @property
    def variableInitialGuesses(self):
        return np.full(len(self.variableNames), DEFAULT_INITIAL_CONC)

    @property
    def freeCount(self):
        return len(self.freeNames)

    @abstractmethod
    def run(self, totalConcVars):
        pass


class StockTable(Table):
    def __init__(self, master, titration):
        try:
            stockTitles = titration.totalConcentrations.stockTitles
        except AttributeError:
            stockTitles = ("Stock 1", "Stock 2")

        try:
            freeNames = titration.totalConcentrations.freeNames
        except AttributeError:
            freeNames = ("Host", "Guest")

        super().__init__(
            master,
            2,
            0,
            stockTitles,
            maskBlanks=True,
            rowOptions=("titles", "new", "delete"),
            columnOptions=("titles", "new", "delete"),
        )

        self.titration = titration

        self.label(0 - self.headerGridRows, 0, "Stock concentrations:", 4)
        self.label(1 - self.headerGridRows, 2, "Unit:")
        _, self.concsUnit = self.dropdown(
            1 - self.headerGridRows, 3, ("nM", "μM", "mM", "M"), "mM"
        )
        try:
            self.concsUnit.set(titration.totalConcentrations.concsUnit)
        except AttributeError:
            pass

        try:
            self.populate(freeNames, titration.totalConcentrations.stockConcs)
        except AttributeError:
            self.populateDefault(freeNames)

    def populate(self, freeNames, stockConcs):
        for name, row in zip(freeNames, stockConcs):
            self.addRow(
                name, [convertConc(conc, "M", self.concsUnit.get()) for conc in row]
            )

    def populateDefault(self, freeNames):
        for name in freeNames:
            self.addRow(name)


class VolumesTable(Table):
    def __init__(self, master, titration):
        try:
            stockTitles = titration.totalConcentrations.stockTitles
        except AttributeError:
            stockTitles = ("Stock 1", "Stock 2")
        super().__init__(
            master,
            2,
            2,
            stockTitles,
            rowOptions=("readonlyTitles", "delete"),
            columnOptions=(),
        )

        self.titration = titration

        self.label(0 - self.headerGridRows, 0, "Cumulative addition volumes:", 4)
        self.label(1 - self.headerGridRows, 2, "Unit:")
        _, self.volumesUnit = self.dropdown(
            1 - self.headerGridRows, 3, ("nL", "μL", "mL", "L"), "μL"
        )
        try:
            self.volumesUnit.set(titration.totalConcentrations.volumesUnit)
        except AttributeError:
            pass

        self.readonlyEntry(self.headerCells - 1, 1, "Addition title:", align="left")

        if (
            self.titration.totalConcentrations is not None
            and hasattr(titration.totalConcentrations, "volumes")
            and (
                self.titration.totalConcentrations.volumes.shape[0]
                == len(self.titration.additionTitles)
            )
        ):
            self.populate(titration.totalConcentrations.volumes)
        else:
            self.populateDefault()

    # TODO: instead make the columnspan of the titles 2
    def deleteRowButton(self, *args, **kwargs):
        button = super().deleteRowButton(*args, **kwargs)
        button.state(["disabled"])
        return button

    def populate(self, volumes):
        for name, row in zip(self.titration.additionTitles, volumes):
            self.addRow(
                name,
                [
                    self.convertVolume(volume, "L", self.volumesUnit.get())
                    for volume in row
                ],
            )

    def populateDefault(self):
        for name in self.titration.additionTitles:
            self.addRow(name)

    def addColumn(self, firstEntry="", data=None):
        super().addColumn(firstEntry, data)
        column = self.cells.shape[1] - 1
        copyFirstButton = self.button(self.headerCells - 2, column, "Copy first")
        copyFirstButton.configure(
            command=lambda button=copyFirstButton: self.copyFirst(
                button.grid_info()["column"]
            )
        )
        self.cells[self.headerCells - 2, column] = copyFirstButton

        copyTitlesButton = self.button(self.headerCells - 1, column, "Copy from titles")
        copyTitlesButton.configure(
            command=lambda button=copyTitlesButton: self.copyFromTitles(
                button.grid_info()["column"]
            )
        )
        self.cells[self.headerCells - 1, column] = copyTitlesButton

    def copyFirst(self, column):
        cells = self.cells[self.headerCells :, column]
        first = cells[0].get()
        for cell in cells:
            cell.set(first)

    def copyFromTitles(self, column):
        cells = self.cells[self.headerCells :]
        for row in cells:
            title = row[1].get()
            volume = self.getVolumeFromString(title, self.volumesUnit.get())
            if volume is not None:
                row[column].set(volume)

    def getVolumeFromString(self, string, toUnit="L"):
        searchResult = re.search(r"([0-9.]+) ?([nuμm]?)[lL]", string)
        if not searchResult:
            return None
        volume, prefix = searchResult.group(1, 2)
        return self.convertVolume(volume, prefix, toUnit)

    def convertVolume(self, volume, fromUnit, toUnit):
        volume = Decimal(volume)
        convertedVolume = float(
            volume
            * prefixesDecimal[fromUnit.strip("L")]
            / prefixesDecimal[toUnit.strip("L")]
        )
        return f"{convertedVolume:g}"  # strip trailing zeroes


class VolumesPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter volumes")

        height = int(self.master.winfo_height() * 0.8)
        frame = ScrolledFrame(
            self,
            height=height,
            max_width=self.winfo_toplevel().master.winfo_width() - 200,
        )
        frame.pack(expand=True, fill="both")

        innerFrame = frame.display_widget(ttk.Frame, stretch=True)

        unknownConcsFrame = ttk.Frame(innerFrame, borderwidth=5)
        unknownConcsFrame.pack(expand=True, fill="both")
        unknownConcsLabel = WrappedLabel(
            unknownConcsFrame,
            text="Leave cells blank to optimise that concentration as a variable.\n",
        )
        unknownConcsLabel.pack(expand=False, fill="both")
        self.unknownTotalConcsLinkedVar = tk.BooleanVar()
        try:
            self.unknownTotalConcsLinkedVar.set(
                self.titration.totalConcentrations.unknownTotalConcsLinked
            )
        except AttributeError:
            self.unknownTotalConcsLinkedVar.set(True)
        unknownTotalConcsCheckbutton = ttk.Checkbutton(
            unknownConcsFrame,
            variable=self.unknownTotalConcsLinkedVar,
            text="Link unknown concentrations in the same row?",
        )
        unknownTotalConcsCheckbutton.pack()

        self.stockTable = StockTable(innerFrame, titration)
        self.stockTable.pack(expand=True, fill="both")
        self.volumesTable = VolumesTable(innerFrame, titration)
        self.volumesTable.pack(expand=True, fill="both")

        self.stockTable.newColumnButton.configure(command=self.addColumns)
        self.stockTable._deleteColumn = self.stockTable.deleteColumn
        self.stockTable.deleteColumn = self.deleteColumns

        buttonFrame = ButtonFrame(innerFrame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both")

    def addColumns(self):
        self.stockTable.addColumn()
        self.volumesTable.addColumn()

    def deleteColumns(self, column):
        self.stockTable._deleteColumn(column)
        self.volumesTable.deleteColumn(column)

    def reset(self):
        for table in (self.stockTable, self.volumesTable):
            table.resetData()
            table.columnTitles = ("Stock 1", "Stock 2")
            table.populateDefault()

    def saveData(self):
        self.freeNames = self.stockTable.rowTitles

        self.stockTitles = self.stockTable.columnTitles
        self.unknownTotalConcsLinked = self.unknownTotalConcsLinkedVar.get()

        self.concsUnit = self.stockTable.concsUnit.get()
        self.stockConcs = self.stockTable.data * prefixes[self.concsUnit.strip("M")]

        self.volumesUnit = self.volumesTable.volumesUnit.get()
        self.volumes = self.volumesTable.data * prefixes[self.volumesUnit.strip("L")]

        self.saved = True
        self.destroy()


class GetTotalConcsFromVolumes(totalConcentrations):
    Popup = VolumesPopup
    popupAttributes = (
        "stockTitles",
        "unknownTotalConcsLinked",
        "concsUnit",
        "stockConcs",
        "volumesUnit",
        "volumes",
        "freeNames",
    )

    def run(self, totalConcVars):
        stockConcs = np.copy(self.stockConcs)
        if self.unknownTotalConcsLinked:
            # For each row (= species), all blank cells are assigned to a
            # single unknown variable.
            for rowIndex, totalConcVar in zip(
                np.where(self.rowsWithBlanks)[0], totalConcVars
            ):
                stockConcs[rowIndex, np.isnan(stockConcs[rowIndex])] = totalConcVar
        else:
            stockConcs[np.isnan(stockConcs)] = totalConcVars

        moles = self.volumes @ stockConcs.T
        totalVolumes = np.atleast_2d(np.sum(self.volumes, 1)).T
        totalConcs = moles / totalVolumes

        return totalConcs

    @property
    def totalConcs(self):
        # If all total concentrations are known, they can be used by other strategies.
        if len(self.variableNames) != 0:
            return np.empty((0, 0))
        else:
            return self.run(np.empty((0,)))

    @property
    def rowsWithBlanks(self):
        return np.any(ma.getmaskarray(self.stockConcs), axis=1)

    @property
    def variableNames(self):
        if self.unknownTotalConcsLinked:
            # return the number of rows (= species) with blank cells
            concVarsNames = self.freeNames[self.rowsWithBlanks]
            return np.array([f"[{name}]" for name in concVarsNames])
        else:
            concVarsNames = []
            for freeName, concs in zip(self.freeNames, self.stockConcs):
                concVarsNames.extend(
                    [
                        f"[{freeName}] in {stock}"
                        for stock in self.stockTitles[ma.getmaskarray(concs)]
                    ]
                )
            return np.array(concVarsNames)


class ConcsTable(Table):
    # TODO: merge with VolumesTable
    def __init__(self, master, titration):
        self.titration = titration

        try:
            freeNames = titration.totalConcentrations.freeNames
        except AttributeError:
            freeNames = ("Host", "Guest")

        super().__init__(
            master,
            2,
            2,
            freeNames,
            rowOptions=("readonlyTitles",),
            columnOptions=("titles", "new", "delete"),
        )

        self.populateDefault()

    def populateDefault(self):
        self.label(0 - self.headerGridRows, 0, "Concentrations:", 4)
        self.label(1 - self.headerGridRows, 2, "Unit:")
        _, self.concsUnit = self.dropdown(
            1 - self.headerGridRows, 3, ("nM", "μM", "mM", "M"), "mM"
        )

        self.readonlyEntry(self.headerCells - 1, 1, "Addition title:", align="left")

        if (
            self.titration.totalConcentrations is not None
            and self.titration.totalConcentrations.totalConcs.shape[0]
            == len(self.titration.additionTitles)
        ):
            self.concsUnit.set(self.titration.totalConcentrations.concsUnit)
            for name, row in zip(
                self.titration.additionTitles,
                self.titration.totalConcentrations.totalConcs,
            ):
                self.addRow(
                    name, [convertConc(conc, "M", self.concsUnit.get()) for conc in row]
                )
        else:
            for name in self.titration.additionTitles:
                self.addRow(name)

    def addColumn(self, firstEntry="", data=None):
        super().addColumn(firstEntry, data)
        column = self.cells.shape[1] - 1
        copyFirstButton = self.button(self.headerCells - 2, column, "Copy first")
        copyFirstButton.configure(
            command=lambda button=copyFirstButton: self.copyFirst(
                button.grid_info()["column"]
            )
        )
        self.cells[self.headerCells - 2, column] = copyFirstButton

        copyTitlesButton = self.button(self.headerCells - 1, column, "Copy from titles")
        copyTitlesButton.configure(
            command=lambda button=copyTitlesButton: self.copyFromTitles(
                button.grid_info()["column"]
            )
        )
        self.cells[self.headerCells - 1, column] = copyTitlesButton

    def copyFirst(self, column):
        cells = self.cells[self.headerCells :, column]
        first = cells[0].get()
        for cell in cells:
            cell.set(first)

    def copyFromTitles(self, column):
        cells = self.cells[self.headerCells :]
        for row in cells:
            title = row[1].get()
            conc = self.getConcFromString(title, self.concsUnit.get())
            if conc is not None:
                row[column].set(conc)

    def getConcFromString(self, string, toUnit="M"):
        searchResult = re.search(r"([0-9.]+) ?([nuμm]?)M", string)
        if not searchResult:
            return None
        conc, prefix = searchResult.group(1, 2)
        return convertConc(conc, prefix, toUnit)


class ConcsPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter concentrations")

        height = int(self.master.winfo_height() * 0.8)
        frame = ScrolledFrame(
            self,
            height=height,
            max_width=self.winfo_toplevel().master.winfo_width() - 200,
        )
        frame.pack(expand=True, fill="both")

        innerFrame = frame.display_widget(ttk.Frame, stretch=True)

        self.concsTable = ConcsTable(innerFrame, titration)
        self.concsTable.pack(expand=True, fill="both")

        buttonFrame = ButtonFrame(innerFrame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both")

    def reset(self):
        self.concsTable.resetData()
        try:
            self.concsTable.columnTitles = self.titration.totalConcentrations.freeNames
        except AttributeError:
            self.concsTable.columnTitles = ("Host", "Guest")
        self.concsTable.populateDefault()

    def saveData(self):
        self.concsUnit = self.concsTable.concsUnit.get()
        self.totalConcs = self.concsTable.data * prefixes[self.concsUnit.strip("M")]
        self.freeNames = self.concsTable.columnTitles

        self.saved = True
        self.destroy()


class GetTotalConcs(totalConcentrations):
    Popup = ConcsPopup
    popupAttributes = (
        "concsUnit",
        "totalConcs",
        "freeNames",
    )

    def run(self, totalConcVars):
        return self.totalConcs

    variableNames = np.array([])


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Experiment"
    dropdownLabelText = "Enter concentrations or volumes:"
    dropdownOptions = {
        "Volumes": GetTotalConcsFromVolumes,
        "Concentrations": GetTotalConcs,
    }
    attributeName = "totalConcentrations"
    setDefault = False



# Based on https://github.com/bmjcode/tkScrolledFrame by
import sys
import time
import tkinter as tk
import tkinter.ttk as ttk

from packaging.version import parse as v_parse

# TODO: replace with custom implementation to solve flickering issues
from ttkwidgets.autohidescrollbar import AutoHideScrollbar


class ScrolledFrame(ttk.Frame):
    fps = 60
    interval = 1 / fps
    steps = int(fps / 10)

    mousewheel_bindings = None

    def bind_mousewheel(self):
        if self.mousewheel_bindings is not None:
            return False
        else:
            events = (
                ("<MouseWheel",)
                if self.use_TIP474
                else ("<MouseWheel>", "<Button-4>", "<Button-5>")
            )
            ScrolledFrame.mousewheel_bindings = [
                self.bind_all(event, self.deliver_event, add=True) for event in events
            ]

    def deliver_event(self, event):
        widget = event.widget
        if type(widget) is str:
            return

        event_name = f"<{event.type._name_}>"
        if any(
            [
                event_name in widget._bind(("bind", tag), None, None, None)
                for tag in widget.bindtags()
                if tag not in ("all", widget.winfo_toplevel()._w)
            ]
        ):
            return  # Widget has its own mouse wheel binding
        # Can't check for whether the widget is the toplevel in the loop condition, as
        # wm_manage could turn a ScrolledFrame into a toplevel window.
        while True:
            if isinstance(widget, ScrolledFrame):
                break
            if widget == widget.winfo_toplevel():
                return  # event occurred outside of any ScrolledFrames
            widget = widget.nametowidget(widget.winfo_parent())
        widget.process_event(event)

    def process_event(self, event):
        orient = "horizontal" if bool(event.state & 1) else "vertical"
        # only scroll if the inner frame doesn't fit fully
        if (
            orient == "horizontal"
            and self._interior.winfo_reqwidth() < self._canvas.winfo_width()
        ) or (
            orient == "vertical"
            and self._interior.winfo_reqheight() < self._canvas.winfo_height()
        ):
            return

        if sys.platform.startswith("darwin") and not self.use_TIP474:
            # macOS, pre Tk 8.7
            self.scroll_canvas(orient, "scroll", -1 * event.delta, "units")

        elif event.num == 4:
            # Unix - scroll up
            self.scroll_canvas(orient, "scroll", -1, "units")

        elif event.num == 5:
            # Unix - scroll down
            self.scroll_canvas(orient, "scroll", 1, "units")

        else:
            # Windows
            self.scroll_canvas(
                orient, "scroll", str(-1 * (event.delta // 120)), "units"
            )

    def scroll_canvas_x(self, *args):
        return self.scroll_canvas("horizontal", *args)

    def scroll_canvas_y(self, *args):
        return self.scroll_canvas("vertical", *args)

    def scroll_canvas(self, orient, *args):
        view = "yview" if orient == "vertical" else "xview"
        if args[0] == "moveto":
            currentTime = time.perf_counter()
            if currentTime - self.lastTime >= self.interval:
                setview = (
                    self._canvas.yview if orient == "vertical" else self._canvas.xview
                )
                setview(*args)
                self.lastTime = currentTime
            else:
                nextTime = self.lastTime + self.interval
                delay = nextTime - currentTime
                if delay <= 2 * self.interval:
                    self.lastTime = nextTime
                    # avoid the overhead of creating a new Python function and calling it
                    # from Tk
                    self.tk.eval(
                        f"after {int(1000 * delay)} [list"
                        f" {self._canvas._w} {view} {' '.join(args)}]"
                    )
        elif args[0] == "scroll" and args[2] == "units":
            for i in range(self.steps):
                totalSize = (
                    self._canvas.winfo_height()
                    if orient == "vertical"
                    else self._canvas.winfo_width()
                )
                distance = int(args[1]) * totalSize / 20 / self.steps
                self.tk.eval(
                    f"after {int(i * 1000 * self.interval)} [list"
                    f" {self._canvas._w} {view} scroll {int(distance * (1 + i))} units]"
                )

    def __init__(self, master=None, autohide=True, max_width=None, **kw):
        """Return a new scrollable frame widget."""
        self.lastTime = 0

        super().__init__(master)

        # Check if we should use the MouseWheel behaviour from TIP 474 introduced in
        # Tk 8.7
        self.use_TIP474 = v_parse(self.tk.getvar("tk_version")) >= v_parse("8.7")

        # Hold these names for the interior widget
        self._interior = None
        self._interior_id = None

        # Which scrollbars to provide
        if "scrollbars" in kw:
            scrollbars = kw["scrollbars"]
            del kw["scrollbars"]

            if not scrollbars:
                scrollbars = self._DEFAULT_SCROLLBARS
            elif scrollbars not in self._VALID_SCROLLBARS:
                raise ValueError(
                    "scrollbars parameter must be one of "
                    "'vertical', 'horizontal', 'both', or "
                    "'neither'"
                )
        else:
            scrollbars = self._DEFAULT_SCROLLBARS

        if autohide:
            Scrollbar = AutoHideScrollbar
        else:
            Scrollbar = ttk.Scrollbar

        self._max_width = max_width

        # Default to a 1px sunken border
        if "borderwidth" not in kw:
            kw["borderwidth"] = 0
        if "relief" not in kw:
            kw["relief"] = "flat"

        # Set up the grid
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        # Canvas to hold the interior widget
        c = self._canvas = tk.Canvas(
            self, borderwidth=0, highlightthickness=0, takefocus=0
        )

        self.bind_mousewheel()

        # Call _resize_interior() when the canvas widget is updated
        c.bind("<Configure>", self._resize_interior)

        # Scrollbars
        xs = self._x_scrollbar = Scrollbar(
            self, orient="horizontal", command=self.scroll_canvas_x
        )
        ys = self._y_scrollbar = Scrollbar(
            self, orient="vertical", command=self.scroll_canvas_y
        )
        c.configure(
            xscrollcommand=xs.set,
            yscrollcommand=ys.set,
            xscrollincrement=1,
            yscrollincrement=1,
        )

        # Lay out our widgets
        c.grid(row=0, column=0, sticky="nsew")
        if scrollbars == "vertical" or scrollbars == "both":
            ys.grid(row=0, column=1, sticky="ns")
        if scrollbars == "horizontal" or scrollbars == "both":
            xs.grid(row=1, column=0, sticky="we")

        # Forward these to the canvas widget
        self.bind = c.bind
        self.focus_set = c.focus_set
        self.unbind = c.unbind
        self.xview = c.xview
        self.xview_moveto = c.xview_moveto
        self.yview = c.yview
        self.yview_moveto = c.yview_moveto

        # Process our remaining configuration options
        self.configure(**kw)

    def __setitem__(self, key, value):
        """Configure resources of a widget."""

        if key in self._CANVAS_KEYS:
            # Forward these to the canvas widget
            self._canvas.configure(**{key: value})

        else:
            # Handle everything else normally
            tk.Frame.configure(self, **{key: value})

    # ------------------------------------------------------------------------

    def cget(self, key):
        """Return the resource value for a KEY given as string."""

        if key in self._CANVAS_KEYS:
            return self._canvas.cget(key)

        else:
            return tk.Frame.cget(self, key)

    # Also override this alias for cget()
    __getitem__ = cget

    def configure(self, cnf=None, **kw):
        """Configure resources of a widget."""

        # This is overridden so we can use our custom __setitem__()
        # to pass certain options directly to the canvas.
        if cnf:
            for key in cnf:
                self[key] = cnf[key]

        for key in kw:
            self[key] = kw[key]

    # Also override this alias for configure()
    config = configure

    def display_widget(self, widget_class, stretch=True, **kw):
        self._stretch = stretch
        """Create and display a new widget.

        If stretch == True, the interior widget will be stretched as
        needed to fit the width of the frame.

        Keyword arguments are passed to the widget_class constructor.

        Returns the new widget.
        """

        # Blank the canvas
        self.erase()

        # Set the new interior widget
        self._interior = widget_class(self._canvas, **kw)

        # Add the interior widget to the canvas, and save its widget ID
        # for use in _resize_interior()
        self._interior_id = self._canvas.create_window(
            0, 0, anchor="nw", window=self._interior
        )

        # Call _update_scroll_region() when the interior widget is resized
        self._interior.bind("<Configure>", self._update_scroll_region)

        # Fit the interior widget to the canvas if requested
        # We don't need to check fit_width here since _resize_interior()
        # already does.
        self._resize_interior()

        # Scroll to the top-left corner of the canvas
        self.scroll_to_top()

        return self._interior

    def erase(self):
        """Erase the displayed widget."""

        # Clear the canvas
        self._canvas.delete("all")

        # Delete the interior widget
        del self._interior
        del self._interior_id

        # Save these names
        self._interior = None
        self._interior_id = None

        # Reset width fitting
        self._fit_width = False

    def scroll_to_top(self):
        """Scroll to the top-left corner of the canvas."""

        self._canvas.xview_moveto(0)
        self._canvas.yview_moveto(0)

    # ------------------------------------------------------------------------

    def _resize_interior(self, event=None):
        """Resize the canvas to fit the interior widget."""
        if self._stretch:
            # The current width of the canvas
            canvas_width = self._canvas.winfo_width()

            # The interior widget's requested width
            requested_width = self._interior.winfo_reqwidth()

            if requested_width != canvas_width:
                # Resize the interior widget
                if self._max_width is not None:
                    requested_width = min(self._max_width, requested_width)
                self._canvas.config(width=requested_width)

            canvas_height = self._canvas.winfo_height()
            requested_height = self._interior.winfo_reqheight()
            if requested_height < canvas_height:
                self._canvas.config(height=requested_height)

    def _update_scroll_region(self, event):
        """Update the scroll region when the interior widget is resized."""

        # The interior widget's requested width and height
        req_width = self._interior.winfo_reqwidth()
        req_height = self._interior.winfo_reqheight()

        # Set the scroll region to fit the interior widget
        self._canvas.configure(scrollregion=(0, 0, req_width, req_height))

    # ------------------------------------------------------------------------

    # Keys for configure() to forward to the canvas widget
    _CANVAS_KEYS = "width", "height", "takefocus"

    # Scrollbar-related configuration
    _DEFAULT_SCROLLBARS = "both"
    _VALID_SCROLLBARS = "vertical", "horizontal", "both", "neither"



import tkinter as tk
import tkinter.ttk as ttk

import numpy as np

from . import moduleFrame
from .scrolledFrame import ScrolledFrame
from .style import padding
from .table import ButtonFrame, WrappedLabel


class ContributingSpecies(moduleFrame.Strategy):
    requiredAttributes = ("filter",)

    # If set by a subclass to an index, then contributors.py will determine the default
    # number of contributing states in each species from the stoichiometry of that
    # molecule.
    singleMoleculeIndex = None

    def mapContributorsToSignals(self, contributorConcs):
        if contributorConcs.ndim == 2:
            return contributorConcs
        elif contributorConcs.ndim == 3:
            return contributorConcs[self.signalToMoleculeMap, :, :]

    def run(self):
        return self.filter


class GetContributingSpeciesSingle(ContributingSpecies):
    requiredAttributes = ContributingSpecies.requiredAttributes + (
        "singleMoleculeIndex",
    )

    @property
    def filter(self):
        free = np.zeros(self.titration.speciation.freeCount, dtype=bool)
        free[self.singleMoleculeIndex] = True

        bound = self.titration.speciation.outputStoichiometries[
            :, self.singleMoleculeIndex
        ].astype(bool)

        return np.concatenate([free, bound])


class GetContributingSpeciesHost(GetContributingSpeciesSingle):
    singleMoleculeIndex = 0


class GetContributingSpeciesAll(ContributingSpecies):
    @property
    def filter(self):
        return np.ones(self.titration.speciation.outputCount, dtype=bool)


class ContributingSpeciesCustomPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Select contributing species")

        height = int(self.master.winfo_height() * 0.4)
        self.frame = ttk.Frame(self, height=height)
        self.frame.pack(expand=True, fill="both")

        label = ttk.Label(
            self.frame,
            text="Select all species that contribute to the signals.",
            justify="left",
            padding=5,
        )
        label.pack(expand=False, fill="x")

        self.checkbuttonsFrame = ttk.Frame(self.frame)
        self.checkbuttonsFrame.pack(expand=False, fill="none")

        self.checkbuttonVars = []
        currentFilter = self.titration.contributingSpecies.filter
        if len(currentFilter.shape) == 2:
            currentFilter = currentFilter.any(axis=0)
        for name, state in zip(
            self.titration.speciation.outputNames,
            self.titration.contributingSpecies.filter,
        ):
            var = tk.BooleanVar(self, value=bool(state))
            self.checkbuttonVars.append(var)
            checkbutton = ttk.Checkbutton(
                self.checkbuttonsFrame, text=name, variable=var
            )
            checkbutton.grid(sticky="w", pady=padding)

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

    def reset(self):
        for var, state in zip(
            self.checkbuttonVars,
            self.titration.contributingSpecies.filter,
        ):
            var.set(bool(state))

    def saveData(self):
        self.filter = np.array([var.get() for var in self.checkbuttonVars])
        self.saved = True
        self.destroy()


class GetContributingSpeciesCustom(ContributingSpecies):
    Popup = ContributingSpeciesCustomPopup
    popupAttributes = ("filter",)


class ContributorsPerSignalNotebook(ttk.Notebook):
    def __init__(self, master, titration, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.titration = titration


class ContributorsPerSignalPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter the contributing species")

        height = int(self.master.winfo_height() * 0.4)
        self.frame = ttk.Frame(self, height=height)
        self.frame.pack(expand=True, fill="both")

        label = WrappedLabel(
            self.frame,
            padding=padding * 2,
            text=(
                "For each signal, specify which molecule (free and in complexes) it is"
                " caused by."
            ),
        )
        label.pack(expand=False, fill="both")

        scrolledFrame = ScrolledFrame(
            self.frame, max_width=self.winfo_toplevel().master.winfo_width() - 200
        )
        scrolledFrame.pack(expand=True, fill="both")
        self.innerFrame = scrolledFrame.display_widget(ttk.Frame, stretch=True)

        self.mapVars = []

        radioFrame = ttk.Frame(self.innerFrame)
        radioFrame.pack(expand=True, fill="both")

        for i, freeName in enumerate(titration.speciation.freeNames):
            label = ttk.Label(radioFrame, text=freeName)
            label.grid(row=0, column=i + 1, sticky="w")
            radioFrame.columnconfigure(i + 1, uniform="map", pad=padding)

        for i, (title, index) in enumerate(
            zip(titration.processedSignalTitlesStrings, self.getDefaultMap())
        ):
            label = ttk.Label(radioFrame, text=title)
            label.grid(row=i + 1, column=0, padx=(3 * padding, int(1.5 * padding)))

            mapVar = tk.IntVar(self, value=index)
            self.mapVars.append(mapVar)
            for j in range(len(titration.speciation.freeNames)):
                radioButton = ttk.Radiobutton(radioFrame, variable=mapVar, value=j)
                radioButton.grid(row=i + 1, column=j + 1, sticky="w")

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

    def getDefaultMap(self):
        if (
            hasattr(self.titration.contributingSpecies, "signalToMoleculeMap")
            and len(self.titration.contributingSpecies.signalToMoleculeMap)
            == self.titration.numSignals
            and np.all(
                self.titration.contributingSpecies.signalToMoleculeMap
                <= self.titration.speciation.freeCount
            )
        ):
            return self.titration.contributingSpecies.signalToMoleculeMap
        else:
            return np.zeros(self.titration.numSignals, dtype=int)

    def reset(self):
        for var, index in zip(self.mapVars, self.getDefaultMap()):
            var.set(index)

    def saveData(self):
        self.signalToMoleculeMap = np.array([var.get() for var in self.mapVars])

        self.saved = True
        self.destroy()


class GetContributingSpeciesPerSignal(ContributingSpecies):
    Popup = ContributorsPerSignalPopup
    popupAttributes = ("signalToMoleculeMap",)

    @property
    def singleMoleculeIndex(self):
        return np.arange(self.titration.speciation.freeCount)

    @property
    def filter(self):
        # rows are all contributing molecules, columns are the corresponding
        # contributing species
        free = np.eye(self.titration.speciation.freeCount, dtype=bool)

        bound = (
            self.titration.speciation.outputStoichiometries[
                :, range(self.titration.speciation.freeCount)
            ]
            .astype(bool)
            .T
        )

        return np.hstack([free, bound])


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Spectra"
    dropdownLabelText = "Which species contribute to the spectra?"
    dropdownOptions = {
        "Only species containing Host": GetContributingSpeciesHost,
        "All species": GetContributingSpeciesAll,
        "Custom": GetContributingSpeciesCustom,
        "Custom, different per signal": GetContributingSpeciesPerSignal,
    }
    attributeName = "contributingSpecies"



import csv
import os
import tkinter as tk
import tkinter.ttk as ttk

import matplotlib as mpl
import numpy as np
from matplotlib.backend_bases import _Mode
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure
from numpy import ma

from .editData import predefinedParams
from .style import padding
from .titration import Titration


def fillPredefinedParams(titration, params):
    for k, v in params._asdict().items():
        setattr(titration, k, v)


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


# TODO: convert to classes, register using ABC
def readUV(filePath):
    titration = Titration()
    # set default parameters for UV-Vis titrations
    fillPredefinedParams(titration, predefinedParams["UV-Vis"])

    with open(filePath, "r", newline="", encoding="utf-8-sig") as inFile:
        reader = csv.reader(inFile)

        titleRow = next(reader)[::2]
        # the title row can contain an extra blank entry, this gets rid of it
        if not titleRow[-1]:
            titleRow.pop(-1)

        wavelengths = []
        absorbances = []
        # skip the column name row
        next(reader)
        for row in reader:
            if not row or not row[0]:
                break
            wavelengths.append(row[0])
            absorbances.append(row[1::2])

    titration.additionTitles = np.array(titleRow)
    titration.signalTitles = np.array(wavelengths)

    # transpose data so that the column is the wavelength
    titration.rawData = np.array(absorbances, dtype=float).T

    return [titration]


class NavigationToolbarHorizontal(NavigationToolbar2Tk):
    def __init__(self, pickPeak, callback, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pickPeak = pickPeak
        self.callback = callback

    def _update_view(self):
        super()._update_view()
        self.callback()

    def press_pan(self, event):
        event.key = "x"
        return super().press_pan(event)

    def drag_pan(self, event):
        event.key = "x"
        return super().drag_pan(event)

    def release_pan(self, event):
        event.key = "x"
        super().release_pan(event)
        self.callback()

    def press_zoom(self, event):
        event.key = "x"
        return super().press_zoom(event)

    def drag_zoom(self, event):
        event.key = "x"
        return super().drag_zoom(event)

    def release_zoom(self, event):
        # check if user clicked without dragging
        shouldPickPeak = not hasattr(self, "lastrect")
        event.key = "x"
        super().release_zoom(event)
        if shouldPickPeak:
            event.zoomClick = True
            self.pickPeak(event)
        self.callback()

    def draw_rubberband(self, event, x0, y0, x1, y1):
        axes = self.canvas.figure.get_axes()
        y0 = axes[-1].bbox.intervaly[0]
        y1 = axes[0].bbox.intervaly[1]
        return super().draw_rubberband(event, x0, y0, x1, y1)


def readNMR(filePath):
    # reads an MNova 1D peaks list
    additionTitles = []
    frequencies = []
    intensities = []
    plotFrequencies = []
    plotIntensities = []

    with open(filePath, "r", newline="", encoding="utf-8-sig") as inFile:
        reader = csv.reader(inFile, delimiter="\t")
        for row in reader:
            if not row or not row[0]:
                break
            additionTitles.append(row[1])
            currentFrequencies = [float(f) for f in row[2::2]]
            currentIntensities = [float(i) for i in row[3::2]]
            numSignals = len(currentFrequencies)
            # TODO: check that file is valid

            frequencies.append(currentFrequencies)
            intensities.append(currentIntensities)
            currentPlotFrequencies = [0]
            currentPlotFrequencies.extend(
                # append each frequency three times to create peak
                f
                for f in currentFrequencies
                for _ in range(3)
            )
            currentPlotFrequencies.append(0)
            plotFrequencies.append(currentPlotFrequencies)

            currentPlotIntensities = [0] * (numSignals * 3 + 2)
            # link the intensity to the middle of the three times it's present
            currentPlotIntensities[2::3] = currentIntensities
            plotIntensities.append(currentPlotIntensities)

        maxF = max(max(f) for f in frequencies)
        minF = min(min(f) for f in frequencies)

        numRows = len(frequencies)
        fig = Figure()
        axList = fig.subplots(
            numRows, 1, sharex=True, sharey=True, gridspec_kw={"hspace": 0, "wspace": 0}
        )
        axList = np.flip(axList)
        axList[0].invert_xaxis()
        for ax, x, y in zip(axList, plotFrequencies, plotIntensities):
            x[0] = maxF + (0.1 * (maxF - minF))
            x[-1] = minF - (0.1 * (maxF - minF))
            ax.plot(x, y, color="black")
            ax.axes.yaxis.set_visible(False)
            ax.set_xlim(x[0], x[-1])
        fig.tight_layout()

        signals = []
        titles = []
        currentSignal = np.full(numRows, None)
        plottedPoints = np.copy(currentSignal)
        cycler = mpl.rcParams["axes.prop_cycle"].by_key()["color"]

        titration = Titration()
        titration.additionTitles = np.array(additionTitles)
        fillPredefinedParams(titration, predefinedParams["NMR"])

        popup = tk.Toplevel()
        popup.title("Pick signals")

        frame = ttk.Frame(popup)
        frame.pack()

        entry = ttk.Entry(frame)
        entry.insert(0, "Signal title")

        flatI = np.concatenate(intensities)
        flatF = np.concatenate(frequencies)

        def rescale():
            maxF, minF = axList[0].get_xlim()
            filteredI = flatI[(minF <= flatF) & (flatF <= maxF)]
            if len(filteredI) > 0:
                ax.set_ylim(0, 1.2 * np.max(filteredI))

        rescale()

        def onClick(e):
            # click outside of plot area
            if e.inaxes is None:
                return
            # zoom/pan click
            if toolbar.mode != _Mode.NONE and not hasattr(e, "zoomClick"):
                return
            i = np.where(axList == e.inaxes)[0][0]

            if e.button == 3:
                # left click
                if plottedPoints[i] is None:
                    return
                plottedPoints[i].remove()
                currentSignal[i] = None
                plottedPoints[i] = None
                canvas.draw()
                canvas.flush_events()
                return

            x = find_nearest(frequencies[i], e.xdata)
            y = intensities[i][frequencies[i].index(x)]
            currentSignal[i] = x
            if plottedPoints[i] is not None:
                # remove previous point
                plottedPoints[i].remove()
                pass

            plottedPoints[i] = e.inaxes.plot(x, y, "o", color=cycler[0])[0]
            canvas.draw()
            canvas.flush_events()

        def next():
            signals.append(np.copy(currentSignal))
            titles.append(entry.get())

            currentSignal.fill(None)
            plottedPoints.fill(None)

            entry.delete(0, "end")
            entry.insert(0, "Signal title")
            cycler.pop(0)

        def save():
            titration.rawData = ma.masked_invalid(np.array(signals, dtype=float).T)
            titration.signalTitles = np.array(titles)
            popup.destroy()

        btn1 = ttk.Button(frame, text="Save signal", command=next)
        btn2 = ttk.Button(frame, text="Submit", style="success.TButton", command=save)
        for widget in (entry, btn1, btn2):
            widget.pack(side="left", padx=2, pady=5)

        canvas = FigureCanvasTkAgg(fig, master=popup)
        canvas.draw()
        canvas.mpl_connect("button_press_event", onClick)
        canvas.get_tk_widget().pack(side="bottom", fill="both", expand=True)

        toolbar = NavigationToolbarHorizontal(
            onClick, rescale, canvas, popup, pack_toolbar=False
        )
        toolbar.update()
        toolbar.pack(side="left", padx=padding)

        for ax in axList:
            # prevent zoom reset when adding points
            ax.autoscale(False)

        popup.wait_window(popup)

        return [titration]


def readMusketeer(filePath):
    serialised = np.load(filePath, allow_pickle=False)
    return [serialised]


fileReaders = np.array(
    [
        ["Musketeer file", "*.fit", readMusketeer],
        ["Cary UV-Vis csv file", "*.csv", readUV],
        ["Mnova NMR peak list", "*.csv", readNMR],
    ]
)



import tkinter.ttk as ttk
from abc import abstractmethod

import numpy as np
from numpy import ma
from scipy.optimize import minimize

from . import moduleFrame
from .scrolledFrame import ScrolledFrame
from .style import padding
from .table import ButtonFrame, Table, WrappedLabel

LN_10 = np.log(10)


def stoichiometriesToBoundNames(freeNames, stoichiometries, polymerMode):
    polymerModes = ("unchanged", "dimer+polymer", "terminal+internal")

    if polymerMode not in polymerModes:
        raise ValueError(
            f"Unknown polymer mode {polymerMode}: must be one of {polymerModes}."
        )
    trans = str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")

    boundNames = []
    for row in stoichiometries:
        boundName = ""
        for freeName, stoichiometry in zip(freeNames, row):
            if stoichiometry == 0:
                continue

            if boundName != "":
                boundName += "·"

            if stoichiometry == 1:
                boundName += freeName
            elif stoichiometry == -1:
                boundName += freeName + "ₙ"
            else:
                boundName += freeName + str(stoichiometry).translate(trans)
        if "ₙ" in boundName and polymerMode == "dimer+polymer":
            # insert a dimer before the polymer entry
            boundNames.append(boundName.replace("ₙ", "₂"))
            boundNames.append(boundName)
        elif "ₙ" in boundName and polymerMode == "terminal+internal":
            # separate the terminal and internal parts of the polymer
            boundNames.append(boundName + " terminal")
            boundNames.append(boundName + " internal")
        else:
            boundNames.append(boundName)
    return np.array(boundNames)


class Speciation(moduleFrame.Strategy):
    requiredAttributes = ("stoichiometries",)

    @abstractmethod
    def run(self, variables, totalConcs):
        pass

    # Other modules need to access freeNames and freeCount, but totalConcentrations
    # may not yet be loaded.
    @property
    def freeNames(self):
        try:
            return self.titration.totalConcentrations.freeNames
        except AttributeError:
            return np.array(["Host", "Guest"])

    @property
    def freeCount(self):
        return len(self.freeNames)

    @property
    def boundCount(self):
        return self.complexCount + 2 * self.polymerCount

    @property
    def polymerIndices(self):
        return np.any(self.stoichiometries < 0, 1)

    @property
    def polymerCount(self):
        return np.count_nonzero(self.stoichiometries < 0)

    @property
    def complexCount(self):
        return self.stoichiometries.shape[0] - self.polymerCount

    @property
    def boundNames(self):
        return stoichiometriesToBoundNames(
            self.freeNames,
            self.stoichiometries,
            "unchanged",
        )

    @property
    def variableStoichiomtries(self):
        # stoichiometries in dimer+polymer mode: insert a dimer before each polymer
        dimerRows = self.stoichiometries[self.polymerIndices].copy()
        dimerRows[dimerRows < 0] = 2
        return np.insert(
            self.stoichiometries, *np.where(self.polymerIndices), dimerRows, axis=0
        )

    @property
    def variableNames(self):
        return stoichiometriesToBoundNames(
            self.freeNames,
            self.stoichiometries,
            "dimer+polymer",
        )

    @property
    def outputStoichiometries(self):
        # stoichiometries in terminal+internal mode: duplicate each polymer
        polymerRows = self.stoichiometries[self.polymerIndices]
        return np.insert(
            self.stoichiometries, *np.where(self.polymerIndices), polymerRows, axis=0
        )

    @property
    def outputNames(self):
        return np.append(
            self.freeNames,
            stoichiometriesToBoundNames(
                self.freeNames,
                self.stoichiometries,
                "terminal+internal",
            ),
        )

    @property
    def outputBoundNames(self):
        return stoichiometriesToBoundNames(
            self.freeNames,
            self.stoichiometries,
            "terminal+internal",
        )

    def variablesToKs(self, variables):
        # The rest of the UI should be agnostic towards the treatment of polymers, so
        # this function is used to separate out the double K values (dimer and 3+_mer)
        # for polymers.
        complexKs = np.empty(self.complexCount)
        polymerK2s = ma.masked_all(self.freeCount)
        polymerKns = ma.masked_all(self.freeCount)

        polymersCounted = 0
        # breakpoint()
        for i in range(self.stoichiometries.shape[0]):
            if self.polymerIndices[i]:
                freeIndex = np.where(self.stoichiometries[i] < 0)[0][0]
                polymerK2s[freeIndex] = variables[i + polymersCounted]
                polymerKns[freeIndex] = variables[i + polymersCounted + 1]
                polymersCounted += 1
            else:
                complexKs[i - polymersCounted] = variables[i + polymersCounted]
        if not len(variables) == self.boundCount:
            raise ValueError(
                f"Expected {self.boundCount * 2} variables, got {len(variables)}."
            )
        return complexKs, polymerK2s, polymerKns

    def freeToBoundConcs(self, freeConcs, complexKs, polymerK2s, polymerKns):
        boundConcs = []
        polymersCounted = 0
        for i in range(self.stoichiometries.shape[0]):
            if self.polymerIndices[i]:
                j = np.where(self.stoichiometries[i] < 0)[0][0]
                terminal = (
                    2
                    * freeConcs[j] ** 2
                    * polymerK2s[j]
                    / (1 - freeConcs[j] * polymerKns[j])
                )
                internal = (
                    freeConcs[j] ** 3
                    * polymerK2s[j]
                    * polymerKns[j]
                    / (1 - freeConcs[j] * polymerKns[j])
                )
                boundConcs.extend([terminal, internal])
                polymersCounted += 1
            else:
                boundConcs.append(
                    complexKs[i - polymersCounted]
                    * np.prod(freeConcs ** self.stoichiometries[i])
                )
        return boundConcs


class SpeciationTable(Table):
    def __init__(self, master, titration):
        sortedTitleLengths = sorted(
            [len(name) for name in titration.totalConcentrations.freeNames]
        )
        self.width = max(6, sortedTitleLengths[-1] + 2)
        self.rowTitleWidth = min(20, sum(sortedTitleLengths[-2:]) + 3 + 2)

        super().__init__(
            master,
            0,
            0,
            titration.totalConcentrations.freeNames,
            rowOptions=("readonlyTitles", "new", "delete"),
            columnOptions=("readonlyTitles",),
            boldTitles=True,
            callback=self.updateTitles,
        )

        if (
            titration.speciation.stoichiometries.shape[1]
            == titration.totalConcentrations.freeCount
        ):
            for boundName, stoichiometry in zip(
                titration.speciation.boundNames, titration.speciation.stoichiometries
            ):
                stoichiometry = stoichiometry.astype(str)
                stoichiometry[stoichiometry == "-1"] = "n"
                self.addRow(boundName, stoichiometry)
        else:
            if titration.totalConcentrations.freeCount == 1:
                self.addRow("", np.array([1]))
            else:
                self.addRow(
                    "",
                    np.concatenate(
                        [
                            [1, 1],
                            [0] * (titration.totalConcentrations.freeCount - 2),
                        ]
                    ).astype(int),
                )
            self.updateTitles()

    def updateTitles(self, *args, **kwargs):
        for cell, title in zip(
            self.cells[2 : self.headerCells :, 1], self.columnTitles
        ):
            cell.set(title)

        self.rowTitles = stoichiometriesToBoundNames(
            self.columnTitles, self.data, polymerMode="unchanged"
        )

    def addFreeRow(self, index=-1):
        row = self.cells.shape[0]
        freeRow = np.full(self.cells.shape[1], None)

        freeRow[0] = self.deleteRowButton(row, 0, "Delete Row", state="disabled")
        freeRow[1] = self.readonlyEntry(
            row, 1, self.columnTitles[index], font=self.titleFont
        )

        freeRow[2:] = [
            self.readonlyEntry(row, 2 + column, "0", align="right")
            for column in range(self.cells.shape[1] - 2)
        ]
        freeRow[-1].set("1")

        self.cells = np.insert(self.cells, self.headerCells, freeRow, axis=0)
        self.headerCells += 1

        self.redraw()

    def addColumn(self, firstEntry="", data=None):
        super().addColumn(firstEntry, data)

        for row in range(2, self.headerCells):
            self.cells[row, -1] = self.readonlyEntry(
                row, self.cells.shape[-1], "0", align="right"
            )

        self.addFreeRow()

    def deleteColumn(self, column):
        super().deleteColumn(column)
        self.headerCells -= 1
        self.deleteRow(column)
        self.updateTitles()

    def convertData(self, number):
        if number == "n":
            return -1
        elif number == "":
            return 0
        else:
            return int(number)


class SpeciationPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter speciation table")

        height = int(self.master.winfo_height() * 0.4)
        self.frame = ttk.Frame(self, height=height)
        self.frame.pack(expand=True, fill="both")

        label = WrappedLabel(
            self.frame,
            padding=padding * 2,
            text=(
                "Each column corresponds to a molecule, and each row to a complex.\n"
                " Define each complex by adding a row with the stoichiometry of each"
                " molecule in the complex. For polymers, use 'n'. Leaving a cell blank"
                " is identical to entering '0'."
            ),
        )
        label.pack(expand=False, fill="both")

        scrolledFrame = ScrolledFrame(
            self.frame, max_width=self.winfo_toplevel().master.winfo_width() - 200
        )
        scrolledFrame.pack(expand=True, fill="both")
        self.innerFrame = scrolledFrame.display_widget(ttk.Frame, stretch=True)

        self.speciationTable = SpeciationTable(self.innerFrame, titration)
        self.speciationTable.pack(expand=True, fill="both")

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="x", side="bottom")

    def reset(self):
        self.speciationTable.destroy()
        self.speciationTable = SpeciationTable(self.innerFrame, self.titration)
        self.speciationTable.pack(expand=True, fill="both")

    def saveData(self):
        self.stoichiometries = self.speciationTable.data

        self.saved = True
        self.destroy()


class SpeciationDimerisation(Speciation):
    freeNames = np.array(["Host"])
    stoichiometries = np.array([[2]])

    def run(self, variables, totalConcs):
        K = variables[0]
        Htot = totalConcs.T[0]
        H = (-1 + np.sqrt(1 + 8 * Htot * K)) / (4 * K)
        H2 = (1 + 4 * Htot * K - np.sqrt(1 + 8 * Htot * K)) / (8 * K)

        free = np.atleast_2d(H).T
        bound = np.atleast_2d(H2).T
        return free, bound


class SpeciationHG(Speciation):
    stoichiometries = np.array([[1, 1]])

    def run(self, variables, totalConcs):
        K = variables[0]
        Htot, Gtot = totalConcs.T
        H = (
            np.sqrt(
                Gtot**2 * K**2 - 2 * Gtot * K * (Htot * K - 1) + (Htot * K + 1) ** 2
            )
            - Gtot * K
            + Htot * K
            - 1
        ) / (2 * K)
        G = (
            np.sqrt(
                Htot**2 * K**2 - 2 * Htot * K * (Gtot * K - 1) + (Gtot * K + 1) ** 2
            )
            - Htot * K
            + Gtot * K
            - 1
        ) / (2 * K)
        HG = H * G * K

        free = np.vstack((H, G)).T
        bound = np.atleast_2d(HG).T
        return free, bound


class SpeciationCOGS(Speciation):
    def COGS(self, M, y, ks, polymerKs):
        free = y.copy()
        bound = np.empty(len(ks))
        polymers = np.any(M < 0, 1)
        # the index of the species making up each of the polymers
        polymerParents = np.nonzero(M[polymers])[1]
        complexes = ~polymers

        P = np.empty(len(ks))
        P[complexes] = np.sum(M[complexes, :], 1)
        P[polymers] = 2 * ks[polymers] * y[polymerParents]
        P = 1 / max(P)
        tol = 1e-7
        while True:
            bound[complexes] = ks[complexes] * np.prod(free ** M[complexes, :], 1)
            # cap the maximum guess to avoid divergence
            bound[polymers] = np.where(
                free[polymerParents] * polymerKs[polymers] >= 1,
                free[polymerParents] ** 2 * polymerKs[polymers] / P,
                (2 - polymerKs[polymers] * free[polymerParents])
                * (ks[polymers] * free[polymerParents] ** 2)
                / ((1 - polymerKs[polymers] * free[polymerParents]) ** 2),
            )
            total = free + abs(M.T) @ bound  # total concentrations of species
            if all((total - y) <= tol * y):
                break
            # to handle 0 total concentration
            invTotal = np.where(total == 0, 1, 1 / total)
            free *= (y * invTotal) ** P

        # For polymers, return separate entries for terminal and internal groups
        bound = []
        paddedPolymerParents = ma.array(np.empty(len(ks)), mask=True)
        paddedPolymerParents[polymers] = polymerParents
        for k, polymerK, stoichiometries, isPolymer, parent in zip(
            ks, polymerKs, M, self.polymerIndices, paddedPolymerParents
        ):
            if not isPolymer:
                bound.append(k * np.prod(free**stoichiometries))
            else:
                # terminal
                bound.append(2 * free[parent] ** 2 * k / (1 - free[parent] * polymerK))
                # internal
                bound.append(
                    (free[parent] ** 3 * k * polymerK)
                    / (1 - free[parent] * polymerK) ** 2
                )

        return free, np.array(bound)

    def run(self, variables, totalConcs):
        ks, polymerKs = self.variablesToKs(variables)
        numPoints = totalConcs.shape[0]
        free = np.zeros((numPoints, self.titration.totalConcentrations.freeCount))
        bound = np.zeros((numPoints, self.complexCount + 2 * self.polymerCount))
        for i in range(numPoints):
            free[i], bound[i] = self.COGS(
                self.stoichiometries, totalConcs[i], ks, polymerKs
            )
        return free, bound


class SpeciationSolver(Speciation):
    # Optimise (total * log10(free))
    # This makes the derivative equal to ln(10) * (free + bound - total) / total
    #
    # So gtol in the minimisation algorithm can be set to the desired precision in
    # the total concentrations.
    def speciationLogConst(self, logFreeConst, ks, k2s, kns, total, M):
        free = 10 ** (logFreeConst / total)
        return (
            np.sum(free)
            - np.sum(logFreeConst) * LN_10
            + ks @ np.prod(free**M, 1)
            + self.speciationPolymers(free, k2s, kns)
        )

    def jacobianCalculatorLogConst(self, logFreeConst, ks, k2s, kns, total, M):
        free = 10 ** (logFreeConst / total)
        return (
            (
                (M.T @ (ks * np.prod(free**M, 1)))
                + free
                + self.jacobianPolymers(free, k2s, kns).filled(0.0)
                - total
            )
            * LN_10
            / total
        )

    def speciationPolymers(self, free, k2s, kns):
        result = np.sum(k2s * free**2 / (1 - free * kns))
        if result is np.ma.masked:
            return 0.0
        else:
            return result

    def jacobianPolymers(self, free, k2s, kns):
        return free**2 * k2s * (2 - free * kns) / (1 - free * kns) ** 2

    def getLowerBoundsLogConst(self, ks, k2s, kns, total, M):
        return total * np.log10(
            (total * self.polymersExactSolution(k2s, kns, total).filled(total))
            / (total + M.T @ (ks * np.prod(total**M, 1)))
        )

    # Could be calculated more precisely, for example as
    # UB = M @ (ks * np.prod(LOWER_BOUND**M, 1)))
    # However, as the gradient at the initial guess is almost always negative, the
    # upper bound is rarely actually used.
    def getUpperBoundsLogConst(self, ks, k2s, kns, total, M):
        upperBoundsPolymers = self.polymersExactSolution(k2s, kns, total)
        return total * np.log10(np.ma.min([total, upperBoundsPolymers], axis=0).data)

    def polymersExactSolution(self, k2s, kns, totals):
        output = np.ma.masked_all(len(totals))
        for i, (k2, kn, total) in enumerate(zip(k2s, kns, totals)):
            if not np.ma.is_masked(k2) and not np.ma.is_masked(kn):
                output[i] = self.polymersExactSolutionSingle(k2, kn, total)
        return output

    def polymersExactSolutionSingle(self, k2, kn, total):
        roots = np.roots(
            [
                k2 * kn - kn**2,
                total * kn**2 + 2 * kn - 2 * k2,
                -1 - 2 * total * kn,
                total,
            ]
        )
        real_roots = np.real(roots[np.isreal(roots)])
        return np.min(real_roots[real_roots > 0])

    def run(self, variables, totalConcs):
        # breakpoint()
        complexKs, polymerK2s, polymerKns = self.variablesToKs(variables)
        numPoints = totalConcs.shape[0]
        M = self.stoichiometries[~self.polymerIndices]
        free = np.empty((numPoints, self.freeCount))
        bound = np.empty((numPoints, self.complexCount + 2 * self.polymerCount))

        for i in range(numPoints):
            additionTotalConcs = totalConcs[i]

            # Filter the to exclude species and complexes that will have a concentration
            # of 0
            zeroTotalConcs = additionTotalConcs == 0
            zeroBound = np.any(self.stoichiometries[:, zeroTotalConcs], 1)
            if all(zeroBound):
                free[i] = additionTotalConcs
                bound[i] = 0
                continue
            zeroComplexes = zeroBound[~self.polymerIndices]

            filteredKs = complexKs[~zeroComplexes]
            filteredTotal = additionTotalConcs[~zeroTotalConcs]
            filteredM = M[~zeroComplexes, :][:, ~zeroTotalConcs]

            filteredK2s = polymerK2s[~zeroTotalConcs]
            filteredKns = polymerKns[~zeroTotalConcs]

            lb = self.getLowerBoundsLogConst(
                filteredKs, filteredK2s, filteredKns, filteredTotal, filteredM
            )
            ub = self.getUpperBoundsLogConst(
                filteredKs, filteredK2s, filteredKns, filteredTotal, filteredM
            )
            if any(lb > ub):
                # Correct for rounding errors. Unsure if this is necessary, as the only
                # obvious case when it should happen is if no complexes are formed,
                # which should be caught above.
                mask = np.logical_and(lb > ub, np.isclose(lb, ub))
                lb[mask], ub[mask] = ub[mask], lb[mask]

            if i == 0:
                # Initial guess: all species 100% free
                x0 = ub
            else:
                # If the total concentration increased, the initial guess is that all
                # the added molecules are free.
                # If the total concentration decreased, the initial guess is that the
                # free concentration decreases by the same fraction.
                #
                # The L-BFGS-B implementation will clip x0 to the bounds if necessary.
                initialGuess = free[i - 1].copy()
                difference = additionTotalConcs - totalConcs[i - 1]
                concsIncreased = difference >= 0

                initialGuess[concsIncreased] += difference[concsIncreased]
                initialGuess[~concsIncreased] *= (
                    totalConcs[i][~concsIncreased] / totalConcs[i - 1][~concsIncreased]
                )
                x0 = filteredTotal * np.log10(initialGuess[~zeroTotalConcs])

            result = minimize(
                self.speciationLogConst,
                x0=x0,
                args=(filteredKs, filteredK2s, filteredKns, filteredTotal, filteredM),
                jac=self.jacobianCalculatorLogConst,
                bounds=np.vstack((lb, ub)).T,
                options={
                    "ftol": 0,
                    "gtol": 1e-7 * LN_10,
                },
            )
            logFree = result.x / filteredTotal

            free[i, ~zeroTotalConcs] = 10**logFree
            free[i, zeroTotalConcs] = 0

            # get the concentrations of the bound species from those of the free
            bound[i] = self.freeToBoundConcs(free[i], complexKs, polymerK2s, polymerKns)

        return free, bound


class SpeciationHG2(SpeciationSolver):
    stoichiometries = np.array([[1, 1], [1, 2]])

    def run(self, variables, totalConcs):
        return super().run(variables, totalConcs)


class SpeciationCustom(SpeciationSolver):
    Popup = SpeciationPopup
    popupAttributes = ("stoichiometries",)


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Equilibria"
    dropdownLabelText = "Select a binding isotherm:"
    dropdownOptions = {
        "1:1 binding": SpeciationHG,
        "1:2 binding": SpeciationHG2,
        "Dimerisation": SpeciationDimerisation,
        "Custom": SpeciationCustom,
        # "Custom Grad": SpeciationCustomGrad,
    }
    attributeName = "speciation"



import os
import tkinter as tk
import tkinter.filedialog as fd
import tkinter.messagebox as mb
import tkinter.ttk as ttk
from copy import deepcopy
from pathlib import PurePath

import matplotlib as mpl
import matplotlib.ticker as mtick
import numpy as np
import packaging.version
import tksheet
from cycler import cycler
from matplotlib.backend_bases import ResizeEvent
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure
from numpy import ma
from scipy.interpolate import make_interp_spline
from ttkbootstrap.widgets import InteractiveNotebook

from . import (
    __version__,
    combineResiduals,
    contributingSpecies,
    contributors,
    editData,
    equilibriumConstants,
    fitSignals,
    knownSignals,
    proportionality,
    speciation,
    totalConcentrations,
)
from .moduleFrame import GroupFrame
from .patchMatplotlib import NavigationToolbarVertical
from .scrolledFrame import ScrolledFrame
from .style import defaultFigureParams, figureParams, padding
from .table import Table
from .titration import Titration, titrationAttributes

# Magic value indicating that the data in a .fit file should be copied from the original
# titration
COPY_ORIGINAL_ARRAY = "COPY_OGIRINAL_ARRAY"

titrationModules = [
    totalConcentrations,
    proportionality,
    speciation,
    equilibriumConstants,
    contributingSpecies,
    contributors,
    knownSignals,
    fitSignals,
    combineResiduals,
]


class TitrationFrame(ttk.Frame):
    def __init__(self, parent, filePath=None, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.filePath = filePath
        self.numFits = 0

        # options bar on the left
        scrolledFrame = ScrolledFrame(self)
        scrolledFrame.grid(column=0, row=0, sticky="nesw")
        self.options = scrolledFrame.display_widget(
            ttk.Frame, stretch=True, padding=(0, 0, padding, 0)
        )

        self.options.groupFrames = {}
        experimentFrame = self.options.groupFrames["Experiment"] = GroupFrame(
            self.options, "Experiment"
        )
        experimentFrame.grid(sticky="nesw", pady=padding)

        editDataButton = ttk.Button(
            experimentFrame,
            text="Edit data",
            command=self.editData,
            style="Outline.TButton",
        )
        editDataButton.pack(fill="x", padx=padding)

        self.moduleFrames = {}
        for mod in titrationModules:
            moduleFrame = mod.ModuleFrame(self.options)
            self.moduleFrames[mod.__name__] = moduleFrame
            moduleFrame.pack(fill="x", padx=padding)

        fitDataButton = ttk.Button(
            self.options, style="success.TButton", text="Fit", command=self.fitData
        )
        fitDataButton.grid(sticky="nesw", pady=padding, ipady=padding)

        separator = ttk.Separator(self.options, orient="horizontal")
        separator.grid(sticky="nesw", pady=padding)

        copyFitButton = ttk.Button(self.options, text="Copy fit", command=self.copyFit)
        copyFitButton.grid(sticky="nesw", pady=padding, ipady=padding)

        self.options.columnconfigure(0, weight=1)

        # tabs with different fits
        self.notebook = InteractiveNotebook(
            self,
            padding=padding,
            newtab=self.newFit,
            style="Flat.Interactive.TNotebook",
        )
        self.notebook.grid(column=1, row=0, sticky="nesw")

    def loadTitration(self, titration):
        if type(titration) is Titration:
            self.originalTitration = titration
            self.newFit()
        elif type(titration) is np.lib.npyio.NpzFile:
            # TODO: move this to titrationReader.py
            self.originalTitration = Titration()
            fileVersion = packaging.version.parse(titration[".version"].item())
            for attribute in titrationAttributes:
                try:
                    data = titration[f".original.{attribute}"]
                except KeyError:
                    continue
                else:
                    if data.shape == ():
                        data = data.item()

                try:
                    mask = titration[f".original.{attribute}.mask"]
                except KeyError:
                    pass
                else:
                    if mask.shape == ():
                        mask = mask.item()
                    data = ma.masked_array(data, mask)
                setattr(self.originalTitration, attribute, data)

            for name in titration[".fits"]:
                fit = Titration()

                for attribute in titrationAttributes:
                    try:
                        data = titration[f"{name}.{attribute}"]
                    except KeyError:
                        continue
                    else:
                        if data.shape == ():
                            data = data.item()

                    if (
                        type(data) is type(COPY_ORIGINAL_ARRAY)
                        and data == COPY_ORIGINAL_ARRAY
                    ):
                        data = deepcopy(getattr(self.originalTitration, attribute))
                    else:
                        try:
                            mask = titration[f"{name}.{attribute}.mask"]
                        except KeyError:
                            pass
                        else:
                            if mask.shape == ():
                                mask = mask.item()
                            data = ma.masked_array(data, mask)
                    setattr(fit, attribute, data)

                for module in titrationModules:
                    moduleFrame = module.ModuleFrame
                    # in case no valid strategy is present in the loaded file
                    setattr(fit, moduleFrame.attributeName, None)

                    try:
                        SelectedStrategy = moduleFrame.dropdownOptions[
                            titration[f"{name}.{moduleFrame.attributeName}"].item()
                        ]
                    except KeyError:
                        # no stategy selected
                        continue

                    selectedStrategy = SelectedStrategy(fit)
                    for popupAttributeName in selectedStrategy.popupAttributes:
                        key = f"{name}.{moduleFrame.attributeName}.{popupAttributeName}"
                        try:
                            data = titration[key]
                        except KeyError:
                            # backwards compatibility:
                            # version 1.2.0 moved freeNames from speciation to
                            #
                            if (
                                fileVersion < packaging.version.parse("1.2.0")
                                and moduleFrame.attributeName == "totalConcentrations"
                                and popupAttributeName == "freeNames"
                            ):
                                if (key := f"{name}.speciation.freeNames") in titration:
                                    # freeNames set in custom speciation
                                    data = titration[key]
                                else:
                                    # could be ["Host"] or ["Host", "Guest"]
                                    if (
                                        key := f"{name}.totalConcentrations.stockConcs"
                                    ) in titration:
                                        freeCount = titration[key].shape[0]
                                    elif (
                                        key := f"{name}.totalConcentrations.totalConcs"
                                    ) in titration:
                                        freeCount = titration[key].shape[1]
                                    else:
                                        continue
                                    data = np.array(["Host", "Guest"][:freeCount])
                            else:
                                continue
                        else:
                            if data.shape == ():
                                data = data.item()

                        try:
                            mask = titration[f"{key}.mask"]
                        except KeyError:
                            pass
                        else:
                            if mask.shape == ():
                                mask = mask.item()
                            data = ma.masked_array(data, mask)
                        setattr(selectedStrategy, popupAttributeName, data)

                    try:
                        selectedStrategy.checkAttributes()
                    except NotImplementedError:
                        # required attribute missing
                        continue
                    setattr(fit, moduleFrame.attributeName, selectedStrategy)

                self.newFit(fit, name, setDefault=False)

            self.numFits = titration[".numFits"].item()

        self.notebook.bind("<<NotebookTabChanged>>", self.switchFit, add=True)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)

    def editData(self):
        root = self.winfo_toplevel()
        popup = editData.EditDataPopup(self.currentTab.titration, master=root)
        popup.geometry(f"+{root.winfo_x()+100}+{root.winfo_y()+100}")
        popup.show()
        self.currentTab.inputSpectraFrame.updateData()
        if (
            len(self.currentTab.tabs()) == 1
            and self.currentTab.inputSpectraFrame.loaded
        ):
            self.currentTab.select(self.currentTab.inputSpectraFrame)

    def updateDpi(self):
        for tab in self.notebook.tabs():
            if isinstance(fitNotebook := self.nametowidget(tab), FitNotebook):
                fitNotebook.updateDpi()

    @property
    def currentTab(self):
        return self.notebook.nametowidget(self.notebook.select())

    def newFit(self, titration=None, name=None, setDefault=True):
        if titration is None:
            titration = deepcopy(self.originalTitration)
        for moduleFrame in self.moduleFrames.values():
            moduleFrame.update(titration, setDefault)
        self.numFits += 1

        fitNotebook = FitNotebook(self, titration)
        if name is None:
            name = f"Fit {self.numFits}"
            titration.title = name

        self.notebook.add(fitNotebook, text=name)
        self.notebook.select(str(fitNotebook))
        fitNotebook.loadTabs()

    def copyFit(self):
        self.newFit(deepcopy(self.currentTab.titration), setDefault=False)

    def switchFit(self, event):
        fitNotebook = self.currentTab
        for moduleFrame in self.moduleFrames.values():
            moduleFrame.update(fitNotebook.titration)

    def fitData(self):
        self.currentTab.fitData()

    def saveFile(self, saveAs=False):
        options = {}
        options[".version"] = __version__
        options[".numFits"] = self.numFits

        for titrationAttribute in titrationAttributes:
            try:
                data = getattr(self.originalTitration, titrationAttribute)
            except AttributeError:
                continue

            if isinstance(data, ma.MaskedArray):
                options[f".original.{titrationAttribute}"] = data.data
                options[f".original.{titrationAttribute}.mask"] = data.mask
            else:
                options[f".original.{titrationAttribute}"] = data

        fits = np.array([])

        for tkpath in self.notebook.tabs():
            tab = self.notebook.nametowidget(tkpath)
            if not isinstance(tab, ttk.Notebook):
                continue
            fit = self.notebook.tab(tkpath)["text"][: -self.notebook._padding_spaces]
            fits = np.append(fits, fit)

            titration = tab.titration
            for titrationAttribute in titrationAttributes:
                if (titrationAttribute == "rawData") and np.array_equal(
                    titration.rawData, self.originalTitration.rawData
                ):
                    options[f"{fit}.{titrationAttribute}"] = COPY_ORIGINAL_ARRAY
                    continue

                try:
                    data = getattr(titration, titrationAttribute)
                except AttributeError:
                    continue

                if isinstance(data, ma.MaskedArray):
                    options[f"{fit}.{titrationAttribute}"] = data.data
                    options[f"{fit}.{titrationAttribute}.mask"] = data.mask
                else:
                    options[f"{fit}.{titrationAttribute}"] = data

            for module in titrationModules:
                moduleFrame = module.ModuleFrame
                strategy = getattr(titration, moduleFrame.attributeName)
                if strategy is None:
                    continue

                options[f"{fit}.{moduleFrame.attributeName}"] = list(
                    moduleFrame.dropdownOptions.keys()
                )[
                    [x.__name__ for x in moduleFrame.dropdownOptions.values()].index(
                        type(getattr(titration, moduleFrame.attributeName)).__name__
                    )
                ]

                for popupAttributeName in strategy.popupAttributes:
                    key = f"{fit}.{moduleFrame.attributeName}.{popupAttributeName}"
                    data = getattr(strategy, popupAttributeName)

                    if isinstance(data, ma.MaskedArray):
                        options[key] = data.data
                        options[f"{key}.mask"] = data.mask
                    else:
                        options[key] = data

        options[".fits"] = fits

        if self.filePath is None:
            filePath = fd.asksaveasfilename(
                title="Save as",
                initialfile="New Titration.fit",
                filetypes=[("Musketeer file", "*.fit")],
                defaultextension=".fit",
            )
        elif saveAs or PurePath(self.filePath).suffix != ".fit":
            filePath = fd.asksaveasfilename(
                title="Save as",
                initialdir=PurePath(self.filePath).parent,
                initialfile=PurePath(self.filePath).stem + ".fit",
                filetypes=[("Musketeer file", "*.fit")],
                defaultextension=".fit",
            )
        else:
            filePath = self.filePath

        if filePath != "":
            try:
                with open(filePath, "wb") as f:
                    np.savez_compressed(f, **options)
            except Exception as e:
                mb.showerror(title="Failed to save file", message=e, parent=self)
            else:
                if filePath != self.filePath:
                    self.filePath = filePath
                    self.master.tab(self, text=PurePath(self.filePath).name)


class FitNotebook(ttk.Notebook):
    def __init__(self, master, titration, *args, **kwargs):
        super().__init__(master, padding=padding, style="Flat.TNotebook")
        self.titration = titration

    def add(self, tab, *args, hidden=False, **kwargs):
        super().add(tab, *args, **kwargs)
        if hidden:
            self.hide(tab)
        self.update()

    def loadTabs(self):
        self.inputSpectraFrame = InputSpectraFrame(self, self.titration)
        self.add(self.inputSpectraFrame, text="Input Spectra", hidden=True)
        self.inputSpectraFrame.updateData()

        if hasattr(self.titration, "fitResult"):
            try:
                self.showFit()
            except Exception as e:
                print("Warning: failed to load previous fit result:", e)

    def fitData(self):
        self.tk.eval("tk busy .")
        self.update()
        try:
            self.titration.fitData(self.fitCallback)
        except Exception as e:
            mb.showerror(title="Failed to fit data", message=e, parent=self)
            return
        finally:
            self.tk.eval("tk busy forget .")
            self.update()
        self.showFit()

    def fitCallback(self, *args):
        self.update()

    def showFit(self):
        if hasattr(self.titration, "fitResult"):
            lastTabClass = type(self.nametowidget(self.select()))
            for tab in self.tabs():
                widget = self.nametowidget(tab)
                if isinstance(widget, InputSpectraFrame):
                    continue
                self.forget(widget)
                widget.destroy()
        else:
            lastTabClass = None

        if self.titration.continuous:
            self.continuousFittedFrame = ContinuousFittedFrame(self, self.titration)
            self.add(self.continuousFittedFrame, text="Fitted Spectra")
            self.discreteFittedFrame = DiscreteFromContinuousFittedFrame(
                self, self.titration
            )
            self.add(
                self.discreteFittedFrame,
                text=f"Fit at select {self.titration.xQuantity}",
            )
        else:
            self.discreteFittedFrame = DiscreteFittedFrame(self, self.titration)
            self.add(self.discreteFittedFrame, text="Fitted signals")

        self.speciationFrame = SpeciationFrame(self, self.titration)
        self.add(self.speciationFrame, text="Speciation")

        self.resultsFrame = ResultsFrame(self, self.titration)
        self.add(self.resultsFrame, text="Results")

        if lastTabClass is not None:
            for tab in self.tabs():
                widget = self.nametowidget(tab)
                if isinstance(widget, lastTabClass):
                    self.select(str(widget))
                    break
        else:
            self.select(str(self.discreteFittedFrame))

    def updateDpi(self):
        for tab in self.tabs():
            if isinstance(plotFrame := self.nametowidget(tab), PlotFrame):
                plotFrame.updateDpi()


class PlotFrame(ttk.Frame):
    @property
    def dpi(self):
        return figureParams["x"] / self.figwidth

    @property
    def figwidth(self):
        return defaultFigureParams["x"] * 80 / figureParams["scale"] / 100

    @property
    def figheight(self):
        return defaultFigureParams["y"] * 80 / figureParams["scale"] / 100

    @property
    def canvaswidth(self):
        return figureParams["x"]

    @property
    def canvasheight(self):
        return figureParams["y"]

    def updateDpi(self):
        self.fig.set_size_inches(self.figwidth, self.figheight)
        mpl.rcParams["savefig.dpi"] = self.dpi
        self.fig.canvas.get_tk_widget().configure(
            width=self.canvaswidth, height=self.canvasheight
        )
        self.fig.canvas.updateDpi(
            self.fig.canvas.get_tk_widget().winfo_width(),
            self.fig.canvas.get_tk_widget().winfo_height(),
        )
        self.fig.canvas.draw_idle()


class FigureCanvasTkAggFixedRatio(FigureCanvasTkAgg):
    @property
    def dpi(self):
        return figureParams["x"] / self.figwidth

    @property
    def figwidth(self):
        return defaultFigureParams["x"] / 100 * 80 / figureParams["scale"]

    @property
    def figheight(self):
        return defaultFigureParams["y"] / 100 * 80 / figureParams["scale"]

    @property
    def desiredCanvaswidth(self):
        return figureParams["x"]

    @property
    def desiredCanvasheight(self):
        return figureParams["y"]

    def updateDpi(self, width, height):
        if width == 1 and height == 1:
            return width, height

        if width == self.desiredCanvaswidth and height == self.desiredCanvasheight:
            self.figure.set_dpi(self.dpi)
        else:
            desiredRatio = self.desiredCanvasheight / self.desiredCanvaswidth
            actualRatio = height / width
            if actualRatio > desiredRatio:
                height = width * desiredRatio
                self.figure.set_dpi(height / self.figheight)
            elif actualRatio < desiredRatio:
                width = height / desiredRatio
                self.figure.set_dpi(width / self.figwidth)

        return width, height

    def resize(self, event):
        width, height = event.width, event.height

        width, height = self.updateDpi(width, height)

        self.get_tk_widget().delete(self._tkphoto)
        self._tkphoto = tk.PhotoImage(
            master=self.get_tk_widget(), width=int(width), height=int(height)
        )
        self.get_tk_widget().create_image(
            int(width / 2), int(height / 2), image=self._tkphoto
        )
        ResizeEvent("resize_event", self)._process()
        self.draw_idle()


class InputSpectraFrame(PlotFrame):
    def __init__(self, parent, titration, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration
        self.populate()

        self.loaded = False

    def updateData(self):
        if self.titration.continuous:
            if not self.loaded:
                self.master.add(self, text="Input Spectra")
                self.loaded = True
            self.updateRangeSelection()
            self.plot()
        else:
            self.loaded = False
            self.master.hide(self)

    def populate(self):
        rangeSelection = ttk.Frame(self)
        rangeSelection.grid(row=0, column=1, sticky="s")

        self.rangeLabel = ttk.Label(rangeSelection)
        self.rangeLabel.pack(side="left")

        self.fromSpinbox = ttk.Spinbox(rangeSelection, width=5)
        self.fromSpinbox.pack(padx=padding, side="left")

        ttk.Label(rangeSelection, text="to").pack(side="left")

        self.toSpinbox = ttk.Spinbox(rangeSelection, width=5)
        self.toSpinbox.pack(padx=padding, side="left")

        self.fig = Figure(
            layout="constrained", figsize=(self.figwidth, self.figheight), dpi=self.dpi
        )
        self.ax = self.fig.add_subplot()

        ttk.Button(rangeSelection, text="Update", command=self.updateWLRange).pack(
            side="left"
        )

        canvas = FigureCanvasTkAggFixedRatio(self.fig, master=self)
        canvas.draw()
        canvas.get_tk_widget().grid(row=1, column=1, sticky="nw")

        toolbar = NavigationToolbarVertical(canvas, self, pack_toolbar=False)
        toolbar.update()
        toolbar.grid(row=1, column=0, sticky="ne")

        self.columnconfigure(
            0,
            weight=1000,
            uniform="column",
            minsize=max(w.winfo_reqwidth() for w in toolbar.children.values()),
        )
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1000, uniform="column")
        self.rowconfigure(
            0,
            weight=1000,
            uniform="row",
            minsize=max(w.winfo_reqheight() for w in rangeSelection.children.values()),
        )
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1000, uniform="row")
        self.grid_anchor("center")

    def updateRangeSelection(self):
        self.rangeLabel.configure(text=f"Range of {self.titration.xQuantity} to fit:")

        signalMin = self.titration.signalTitles.min()
        signalMax = self.titration.signalTitles.max()

        currentMin, currentMax = self.titration.continuousRange
        currentMin = max(currentMin, signalMin)
        currentMax = min(currentMax, signalMax)

        decimals = self.titration.signalTitlesDecimals
        step = 1 / (10**decimals)

        self.fromSpinbox.configure(from_=signalMin, to=signalMax, increment=step)
        self.fromSpinbox.set(f"{currentMin:.{decimals}f}")
        self.toSpinbox.configure(from_=signalMin, to=signalMax, increment=step)
        self.toSpinbox.set(f"{currentMax:.{decimals}f}")

    def plot(self):
        ax = self.ax
        fig = self.fig
        titration = self.titration

        ax.cla()

        spectraColors = (
            ["black"] + ["#80808080"] * (titration.numAdditions - 2) + ["tab:red"]
        )
        colorCycler = cycler(color=spectraColors)
        ax.set_prop_cycle(colorCycler)

        ax.plot(titration.processedSignalTitles, titration.processedData.T)

        ax.set_xlabel(f"{titration.xQuantity} / {titration.xUnit}")
        ax.set_ylabel(f"{titration.yQuantity} / {titration.yUnit}")

        fig.canvas.draw_idle()

    def updateWLRange(self):
        from_ = float(self.fromSpinbox.get())
        to = float(self.toSpinbox.get())
        self.titration.continuousRange = np.array([from_, to])
        self.plot()


class ContinuousFittedFrame(PlotFrame):
    def __init__(self, parent, titration, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration
        self.populate()
        self.plot()

    def populate(self):
        self.fig = Figure(
            layout="constrained", figsize=(self.figwidth, self.figheight), dpi=self.dpi
        )
        self.ax = self.fig.add_subplot()
        self.canvas = FigureCanvasTkAggFixedRatio(self.fig, master=self)
        self.canvas.draw()
        self.canvas.get_tk_widget().grid(row=1, column=1, sticky="nw")

        toolbar = NavigationToolbarVertical(self.canvas, self, pack_toolbar=False)
        toolbar.update()
        toolbar.grid(row=1, column=0, sticky="ne")

        self.optionsFrame = ttk.Frame(self)
        self.optionsFrame.grid(row=1, column=2, sticky="w")
        self.speciesLabel = ttk.Label(
            self.optionsFrame, anchor="center", justify="center", text="Show:"
        )
        self.speciesLabel.pack(pady=padding, fill="x")

        self.deconvolutionVar = tk.StringVar()
        deconvolutionOptions = (
            f"Molar {self.titration.yQuantity}",
            "Deconvolution at start",
            "Deconvolution at endpoint",
        )
        self.deconvolutionDropdown = ttk.OptionMenu(
            self.optionsFrame,
            self.deconvolutionVar,
            deconvolutionOptions[0],
            *deconvolutionOptions,
            command=lambda *args: self.plot(),
            style="primary.Outline.TMenubutton",
        )
        self.deconvolutionDropdown.pack()

        self.columnconfigure(
            0,
            weight=1000,
            uniform="column",
            minsize=max(w.winfo_reqwidth() for w in toolbar.children.values()),
        )
        self.columnconfigure(1, weight=1, minsize=0)
        self.columnconfigure(
            2,
            weight=1000,
            uniform="column",
            minsize=max(
                w.winfo_reqwidth() for w in self.optionsFrame.children.values()
            ),
        )
        self.rowconfigure(0, weight=1000, uniform="row")
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1000, uniform="row")
        self.grid_anchor("center")

    def plot(self):
        self.ax.clear()

        titration = self.titration
        spectra = titration.lastFittedSpectra
        names = titration.contributors.outputNames
        wavelengths = titration.processedSignalTitles

        deconvolution = self.deconvolutionVar.get()
        if deconvolution == f"Molar {titration.yQuantity}":
            for spectrum, name in zip(spectra, names):
                self.ax.plot(wavelengths, spectrum, label=name)
            self.ax.set_ylabel(f"molar {titration.yQuantity} / {titration.yUnit} M⁻¹")
        else:
            if deconvolution == "Deconvolution at start":
                deconvolutionPoint = 0
            elif deconvolution == "Deconvolution at endpoint":
                deconvolutionPoint = -1
            else:
                raise ValueError(f"Unknown deconvolution option {deconvolution}")

            self.ax.plot(
                wavelengths,
                titration.processedData[deconvolutionPoint],
                color="0.5",
                label="Observed",
            )
            self.ax.plot(
                wavelengths,
                titration.lastFittedCurves[deconvolutionPoint],
                color="0",
                linestyle="--",
                label="Fitted",
            )
            for spectrum, name in zip(
                spectra * titration.lastSignalVars[[deconvolutionPoint], :].T, names
            ):
                self.ax.plot(
                    wavelengths,
                    spectrum,
                    label=name,
                )
            self.ax.set_ylabel(f"{titration.yQuantity} / {titration.yUnit}")

        self.ax.legend(draggable=True)
        self.ax.set_xlabel(f"{titration.xQuantity} / {titration.xUnit}")
        self.canvas.draw()


class FittedFrame(PlotFrame):
    def __init__(self, parent, titration, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration
        self.xQuantity = titration.totalConcentrations.freeNames[-1]
        self.xConcs = titration.lastTotalConcs.T[-1]
        self.smooth = True
        self.logScale = False

    def populate(self):
        self.fig = Figure(
            layout="constrained", figsize=(self.figwidth, self.figheight), dpi=self.dpi
        )
        self.ax = self.fig.add_subplot()

        self.canvas = FigureCanvasTkAggFixedRatio(self.fig, master=self)
        self.canvas.draw()
        self.canvas.get_tk_widget().grid(row=1, column=1, sticky="nw")

        self.toolbar = NavigationToolbarVertical(self.canvas, self, pack_toolbar=False)
        self.toolbar.update()
        self.toolbar.grid(row=1, column=0, sticky="ne")

        self.toggleButtonsFrame = ttk.Frame(self)
        self.toggleButtonsFrame.grid(row=1, column=2, sticky="w")

        self.plotTypeVar = tk.StringVar()
        self.plotTypes = {
            f"Absolute {self.titration.yQuantity}": "absolute",
            f"Change in {self.titration.yQuantity}": "relative",
            f"Normalised change in {self.titration.yQuantity}": "normalised",
        }
        self.plotTypeDropdown = ttk.OptionMenu(
            self.toggleButtonsFrame,
            self.plotTypeVar,
            [k for (k, v) in self.plotTypes.items() if v == "relative"][0],
            *self.plotTypes.keys(),
            command=lambda *args: self.plot(),
            style="primary.Outline.TMenubutton",
        )
        self.plotTypeDropdown.pack()

        self.logScaleButton = ttk.Checkbutton(
            self.toggleButtonsFrame,
            text="Logarithmic x axis",
            command=self.toggleLogScale,
            style="Outline.Toolbutton",
        )
        self.logScaleButton.pack(pady=padding, fill="x")
        self.smoothButton = ttk.Checkbutton(
            self.toggleButtonsFrame,
            text="Smooth curves",
            command=self.toggleSmooth,
            style="Outline.Toolbutton",
        )
        self.smoothButton.state(("selected",))
        self.smoothButton.pack(pady=padding, fill="x")

        separator = ttk.Separator(self.toggleButtonsFrame, orient="horizontal")
        separator.pack(pady=padding, fill="x")

        self.saveCurvesButton = ttk.Button(
            self.toggleButtonsFrame,
            text="Save fitted curves",
            command=self.saveCurves,
            style="success.Outline.TButton",
        )
        self.saveCurvesButton.pack(pady=padding, fill="x")

        self.columnconfigure(
            0,
            weight=1000,
            uniform="column",
            minsize=max(w.winfo_reqwidth() for w in self.toolbar.children.values()),
        )
        self.columnconfigure(1, weight=1, minsize=0)
        self.columnconfigure(
            2,
            weight=1000,
            uniform="column",
            minsize=max(
                w.winfo_reqwidth() for w in self.toggleButtonsFrame.children.values()
            ),
        )
        self.rowconfigure(0, weight=1000, uniform="row")
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1000, uniform="row")
        self.grid_anchor("center")

    @property
    def plotType(self):
        return self.plotTypes[self.plotTypeVar.get()]

    def toggleLogScale(self):
        self.logScale = not self.logScale
        if self.logScale:
            self.ax.set_xscale("log")
        else:
            self.ax.set_xscale("linear")
        self.canvas.draw()

    def toggleSmooth(self):
        self.smooth = not self.smooth
        self.plot()

    def saveCurves(self):
        initialfile = os.path.splitext(self.titration.title)[0] + "_fitted_curves"
        fileName = fd.asksaveasfilename(
            title="Save fitted curves",
            initialfile=initialfile,
            filetypes=[("CSV file", "*.csv")],
            defaultextension=".csv",
        )
        data = self.titration.lastFittedCurves
        rowTitles = np.atleast_2d(self.titration.additionTitles).T
        columnTitles = np.append("", self.titration.processedSignalTitles)
        output = np.vstack((columnTitles, np.hstack((rowTitles, data))))
        try:
            np.savetxt(fileName, output, fmt="%s", delimiter=",", encoding="utf-8-sig")
        except Exception as e:
            mb.showerror(title="Could not save file", message=e, parent=self)

    def plot(self):
        self.ax.clear()

        titration = self.titration

        # xQuantity and xUnit for the fitted plot. Different from the xQuantity
        # and xUnit in the titration object, which are used for the input
        # spectra.
        xQuantity = self.xQuantity
        xUnit = titration.totalConcentrations.concsUnit
        xConcs = self.xConcs / totalConcentrations.prefixes[xUnit.strip("M")]

        if self.plotType == "absolute":
            curves = self.curves
            fittedCurves = self.fittedCurves
            self.ax.set_ylabel(f"{titration.yQuantity} / {titration.yUnit}")
        elif self.plotType == "relative":
            fittedZeros = self.fittedCurves[:, [0]]
            curves = self.curves - fittedZeros
            fittedCurves = self.fittedCurves - fittedZeros
            self.ax.set_ylabel(f"Δ{titration.yQuantity} / {titration.yUnit}")
        elif self.plotType == "normalised":
            curves = self.curves.T
            fittedCurves = self.fittedCurves.T
            # normalise so that all the fitted curves have the same amplitude
            fittedDiff = self.fittedCurves.T - self.fittedCurves.T[0]
            maxFittedDiff = np.max(abs(fittedDiff), axis=0)
            curves = curves / maxFittedDiff

            diff = curves - curves[0]
            negatives = abs(np.amin(diff, axis=0)) > abs(np.amax(diff, axis=0))
            curves[:, negatives] *= -1
            curves = curves.T * 100

            fittedCurves = fittedCurves / maxFittedDiff
            fittedCurves[:, negatives] *= -1
            fittedCurves = fittedCurves.T * 100

            fittedZeros = fittedCurves[:, [0]]
            curves = curves - fittedZeros
            fittedCurves = fittedCurves - fittedZeros
            self.ax.set_ylabel(f"Normalised Δ{titration.yQuantity} / %")
        else:
            raise ValueError(f"Unknown plot type: {self.plotType}")

        for curve, fittedCurve, name in zip(curves, fittedCurves, self.names):
            self.ax.scatter(xConcs, curve)

            # add step - 1 points between each data point
            step = 10
            smoothXCount = (xConcs.size - 1) * step + 1
            smoothX = np.interp(
                np.arange(smoothXCount), np.arange(smoothXCount, step=step), xConcs
            )

            # make_interp_spline requires all x values to be unique
            filter = np.concatenate((np.diff(xConcs).astype(bool), [True]))
            if (not self.smooth) or xConcs[filter].size < 3:
                # cannot do spline interpolation with fewer than 3 unique x-values
                self.ax.plot(xConcs, fittedCurve, label=name)
                continue
            try:
                spl = make_interp_spline(
                    xConcs[filter], fittedCurve[filter], bc_type="natural"
                )
                smoothY = spl(smoothX)
                self.ax.plot(smoothX, smoothY, label=name)
            except ValueError:
                # spline interpolation failed
                self.ax.plot(xConcs, fittedCurve, label=name)
                continue

        if self.logScale:
            self.ax.set_xscale("log")
        else:
            self.ax.set_xscale("linear")
        self.ax.set_xlabel(f"[{xQuantity}] / {xUnit}")
        self.ax.legend(draggable=True)

        self.canvas.draw()


class DiscreteFittedFrame(FittedFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.curves = self.titration.processedData.T.copy()
        self.fittedCurves = self.titration.lastFittedCurves.T.copy()
        self.names = self.titration.processedSignalTitlesStrings
        self.populate()
        self.plot()


class DiscreteFromContinuousFittedFrame(FittedFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        peakIndices = self.titration.getPeakIndices()
        self.curves = self.titration.processedData.T[peakIndices].copy()
        self.fittedCurves = self.titration.lastFittedCurves.T[peakIndices].copy()
        peakTitles = self.titration.processedSignalTitlesStrings[peakIndices]
        self.names = [f"{title} {self.titration.xUnit}" for title in peakTitles]
        self.populate()
        self.plot()


class SpeciationFrame(PlotFrame):
    def __init__(self, parent, titration, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration
        self.xQuantity = titration.totalConcentrations.freeNames[-1]
        self.xConcs = titration.lastTotalConcs.T[-1]
        self.speciesVar = tk.StringVar(self)
        self.logScale = False
        self.smooth = True
        self.populate()
        self.plot()

    def populate(self):
        titration = self.titration
        self.fig = Figure(
            layout="constrained", figsize=(self.figwidth, self.figheight), dpi=self.dpi
        )
        self.ax = self.fig.add_subplot()

        self.canvas = FigureCanvasTkAggFixedRatio(self.fig, master=self)
        self.canvas.draw()
        self.canvas.get_tk_widget().grid(row=1, column=1, sticky="nw")

        self.toolbar = NavigationToolbarVertical(self.canvas, self, pack_toolbar=False)
        self.toolbar.update()
        self.toolbar.grid(row=1, column=0, sticky="ne")

        self.optionsFrame = ttk.Frame(self)
        self.optionsFrame.grid(row=1, column=2, sticky="w")
        self.speciesLabel = ttk.Label(
            self.optionsFrame, anchor="center", justify="center", text="Select species:"
        )
        self.speciesLabel.pack(pady=padding, fill="x")

        self.speciesDropdown = ttk.OptionMenu(
            self.optionsFrame,
            self.speciesVar,
            titration.totalConcentrations.freeNames[0],
            command=lambda *args: self.plot(),
            *titration.totalConcentrations.freeNames,
            style="primary.Outline.TMenubutton",
        )
        self.speciesDropdown.pack()
        self.logScaleButton = ttk.Checkbutton(
            self.optionsFrame,
            text="Logarithmic x axis",
            command=self.toggleLogScale,
            style="Outline.Toolbutton",
        )
        self.logScaleButton.pack(pady=padding, fill="x")
        self.smoothButton = ttk.Checkbutton(
            self.optionsFrame,
            text="Smooth curves",
            command=self.toggleSmooth,
            style="Outline.Toolbutton",
        )
        self.smoothButton.state(("selected",))
        self.smoothButton.pack(pady=padding, fill="x")

        self.columnconfigure(
            0,
            weight=1000,
            uniform="column",
            minsize=max(w.winfo_reqwidth() for w in self.toolbar.children.values()),
        )
        self.columnconfigure(1, weight=1, minsize=0)
        self.columnconfigure(
            2,
            weight=1000,
            uniform="column",
            minsize=max(
                w.winfo_reqwidth() for w in self.optionsFrame.children.values()
            ),
        )
        self.rowconfigure(0, weight=1000, uniform="row")
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1000, uniform="row")
        self.grid_anchor("center")

    def toggleLogScale(self):
        self.logScale = not self.logScale
        if self.logScale:
            self.ax.set_xscale("log")
        else:
            self.ax.set_xscale("linear")
        self.canvas.draw()

    def toggleSmooth(self):
        self.smooth = not self.smooth
        self.plot()

    @property
    def freeIndex(self):
        return np.where(
            self.titration.totalConcentrations.freeNames == self.speciesVar.get()
        )[0][0]

    def plot(self):
        self.ax.clear()

        titration = self.titration

        # xQuantity and xUnit for the fitted plot. Different from the xQuantity
        # and xUnit in the titration object, which are used for the input
        # spectra.
        totalConcs = titration.lastTotalConcs[:, self.freeIndex]
        additionsFilter = totalConcs != 0
        totalConcs = totalConcs[additionsFilter]

        xQuantity = self.xQuantity
        xUnit = titration.totalConcentrations.concsUnit
        xConcs = (
            self.xConcs[additionsFilter]
            / totalConcentrations.prefixes[xUnit.strip("M")]
        )

        freeConcs = titration.lastFreeConcs[additionsFilter, :][:, self.freeIndex]
        freeName = self.speciesVar.get()

        factor = abs(titration.speciation.outputStoichiometries[:, self.freeIndex])
        boundFilter = factor.astype(bool)

        boundConcs = titration.lastBoundConcs * factor
        boundConcs = boundConcs[additionsFilter, :][:, boundFilter]
        boundNames = titration.speciation.outputBoundNames[boundFilter]

        curves = 100 * np.vstack((freeConcs, boundConcs.T)) / totalConcs
        names = np.append(freeName, boundNames)
        self.ax.set_ylabel(f"% of {freeName}")
        self.ax.yaxis.set_major_formatter(mtick.PercentFormatter(xmax=100))
        self.ax.set_ylim(bottom=-5, top=105)

        for curve, name in zip(curves, names):
            # add step - 1 points between each data point
            step = 10
            smoothXCount = (xConcs.size - 1) * step + 1
            smoothX = np.interp(
                np.arange(smoothXCount), np.arange(smoothXCount, step=step), xConcs
            )

            # make_interp_spline requires all x values to be unique
            filter = np.concatenate((np.diff(xConcs).astype(bool), [True]))
            if (not self.smooth) or xConcs[filter].size < 3:
                # cannot do spline interpolation with fewer than 3 unique x-values
                self.ax.plot(xConcs, curve, label=name)
                continue
            try:
                spl = make_interp_spline(
                    xConcs[filter], curve[filter], bc_type="natural"
                )
                smoothY = spl(smoothX)
                self.ax.plot(smoothX, smoothY, label=name)
            except ValueError:
                # spline interpolation failed
                self.ax.plot(xConcs, curve, label=name)
                continue

        if self.logScale:
            self.ax.set_xscale("log")
        else:
            self.ax.set_xscale("linear")
        self.ax.set_xlabel(f"[{xQuantity}] / {xUnit}")
        self.ax.legend(draggable=True)

        self.canvas.draw()


class ResultsFrame(ttk.Frame):
    def __init__(self, parent, titration, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration
        self.sigfigs = 3
        self.showResults()

    # TODO: add spinbox for sigfigs
    def formatK(self, k):
        return f"{float(f'{k:.{self.sigfigs}g}'):.{max(self.sigfigs, 6)}g}"

    @property
    def bic(self):
        # Bayesian Information Criterion
        numParameters = (
            self.titration.equilibriumConstants.variableCount
            + self.titration.totalConcentrations.variableCount
            + self.titration.contributors.contributorsMatrix.shape[0]
        )
        numDataPoints = self.titration.numAdditions
        chiSquared = np.sum(
            (self.titration.lastFittedCurves - self.titration.processedData) ** 2
            / np.abs(self.titration.lastFittedCurves)
        )

        return numDataPoints * np.log(
            chiSquared / numDataPoints
        ) + numParameters * np.log(numDataPoints)

    def showResults(self):
        titration = self.titration
        bicLabel = ttk.Label(
            self,
            text=f"Bayesian information criterion (lower is better): {self.bic:.3g}",
        )
        bicLabel.pack(side="top", pady=15)

        kTable = Table(
            self,
            0,
            0,
            ["K (M⁻ⁿ)"],
            rowOptions=("readonlyTitles",),
            columnOptions=("readonlyTitles",),
        )

        ks = titration.lastKVars

        for (
            name,
            value,
        ) in zip(titration.equilibriumConstants.variableNames, ks):
            kTable.addRow(name, [self.formatK(value)])
        kTable.pack(side="top", pady=15)

        if titration.totalConcentrations.variableCount > 0:
            concsTable = Table(
                self,
                0,
                0,
                [f"c ({titration.totalConcentrations.concsUnit})"],
                rowOptions=["readonlyTitles"],
                columnOptions=["readonlyTitles"],
            )
            concNames = self.titration.totalConcentrations.variableNames
            concs = titration.lastTotalConcVars
            for concName, conc in zip(concNames, concs):
                concsTable.addRow(
                    concName,
                    [
                        totalConcentrations.convertConc(
                            conc, "M", titration.totalConcentrations.concsUnit
                        )
                    ],
                )
            concsTable.pack(side="top", pady=15)

        sheet = tksheet.Sheet(
            self,
            empty_vertical=0,
            empty_horizontal=0,
            data=list(np.around(titration.lastFittedSpectra, 2)),
            headers=list(titration.processedSignalTitlesStrings),
            row_index=list(titration.contributors.outputNames),
            set_all_heights_and_widths=True,
        )
        sheet.MT.configure(height=0)
        sheet.RI.configure(height=0)

        sheet.enable_bindings()
        sheet.set_width_of_index_to_text()
        sheet.pack(side="top", pady=15, fill="both", expand=True)

        saveButton = ttk.Button(
            self,
            text="Save fitted spectra to CSV",
            command=self.saveCSV,
            style="success.TButton",
        )
        saveButton.pack(side="top", pady=15)

    def saveCSV(self):
        initialfile = os.path.splitext(self.titration.title)[0] + "_fit"
        fileName = fd.asksaveasfilename(
            title="Save fitted spectra",
            initialfile=initialfile,
            filetypes=[("CSV file", "*.csv")],
            defaultextension=".csv",
        )
        data = self.titration.lastFittedSpectra
        rowTitles = np.atleast_2d(self.titration.contributors.outputNames).T
        columnTitles = np.append("", self.titration.processedSignalTitlesStrings)
        output = np.vstack((columnTitles, np.hstack((rowTitles, data))))
        try:
            np.savetxt(fileName, output, fmt="%s", delimiter=",", encoding="utf-8-sig")
        except Exception as e:
            mb.showerror(title="Could not save file", message=e, parent=self)



import tkinter.ttk as ttk
from tkinter import font

import numpy as np
from numpy import ma

from . import moduleFrame
from .scrolledFrame import ScrolledFrame
from .table import ButtonFrame, Table, WrappedLabel

DEFAULT_INITIAL_GUESS = 1000


class EquilibriumConstants(moduleFrame.Strategy):
    requiredAttributes = (
        "kNames",
        "knownKs",
        "initialKs",
    )

    @property
    def initialKs(self):
        return self._initialKs

    @initialKs.setter
    def initialKs(self, value):
        if not isinstance(value, ma.MaskedArray):
            value = ma.masked_invalid(value)
        self._initialKs = value

    @property
    def outputNames(self):
        return self.titration.speciation.variableNames

    @property
    def knownMask(self):
        return ma.getmaskarray(self.knownKs)

    @property
    def variableNames(self):
        return self.kNames[self.knownMask]

    @property
    def variableInitialGuesses(self):
        return self.initialKs[self.knownMask].filled(DEFAULT_INITIAL_GUESS)

    def run(self, kVars):
        ks = self.knownKs.copy()
        ks[self.knownMask] = kVars
        return ks


class GetKsAll(EquilibriumConstants):
    # when every equilibrium constant is unknown and independent
    @property
    def kNames(self):
        return self.titration.speciation.variableNames

    @property
    def knownKs(self):
        return ma.array(np.empty(self.outputCount), mask=True)

    @property
    def initialKs(self):
        return ma.array(np.empty(self.outputCount), mask=True)


# TODO: make trimerIndices work with polymers
class GetKsNonspecific(EquilibriumConstants):
    @property
    def kNames(self):
        return self.titration.speciation.variableNames

    @property
    def knownKs(self):
        knownKs = ma.masked_array(np.empty(self.outputCount))
        knownKs[self.trimerIndices] = 0.001
        knownKs[~self.trimerIndices] = ma.masked
        return knownKs

    @property
    def initialKs(self):
        return ma.array(np.empty(self.outputCount), mask=True)


class CustomKsTable(Table):
    def __init__(self, master, titration):
        self.titration = titration
        self.outputNames = self.titration.speciation.variableNames

        if hasattr(titration.equilibriumConstants, "ksMatrix") and (
            titration.equilibriumConstants.ksMatrix.shape
            == (
                len(titration.equilibriumConstants.kNames),
                len(self.outputNames),
            )
        ):
            kNames = titration.equilibriumConstants.kNames
            ksMatrix = titration.equilibriumConstants.ksMatrix
        else:
            kNames = self.outputNames.copy()
            ksMatrix = np.identity(len(self.outputNames), dtype=int)

        columnTitles = np.append(self.outputNames, "Value")
        self.width = max([len(title) for title in columnTitles] + [14]) + 1

        super().__init__(
            master,
            0,
            0,
            columnTitles,
            allowGuesses=True,
            rowOptions=("titles", "new", "delete"),
            columnOptions=("readonlyTitles",),
            boldTitles=True,
            callback=self.createLabels,
        )
        self.readonlyEntry(
            self.headerCells - 1, 1, "Global K for:", font=self.titleFont
        )
        self.addConstantsRow()
        for name, contributions, knownK, initialK in zip(
            kNames,
            ksMatrix,
            titration.equilibriumConstants.knownKs,
            titration.equilibriumConstants.initialKs,
        ):
            if knownK is not ma.masked:
                value = f"{knownK:g}"
            elif initialK is not ma.masked:
                value = f"~{initialK:g}"
            else:
                value = ""
            self.addRow(name, np.append(contributions, value))

    def newRow(self):
        defaultEntries = np.full(self.dataCells.shape[1], "0")
        defaultEntries[-1] = ""
        self.addRow("New variable", defaultEntries)

    def addConstantsRow(self):
        if (
            hasattr(self.titration.equilibriumConstants, "statisticalFactors")
            and len(self.titration.equilibriumConstants.statisticalFactors)
            == len(self.columnTitles) - 1
        ):
            statisticalFactors = self.titration.equilibriumConstants.statisticalFactors
        else:
            statisticalFactors = np.full(len(self.columnTitles) - 1, "1")
        # Value column doesn't have a statistical factor
        statisticalFactors = np.append(statisticalFactors, "")
        oldRowOptions = self.rowOptions
        self.rowOptions = ("readonlyTitles",)
        self.addRow("Statistical factor", statisticalFactors)
        self.rowOptions = oldRowOptions

        # Value column doesn't have a statistical factor
        self.cells[-1, -1].configure(style="TLabel", takefocus=False)
        self.cells[-1, -1].state(["readonly"])

    def createLabels(self, *args, **kwargs):
        try:
            labels = []
            trans = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")
            variables = self.rowTitles[1:]
            for globalK, statFactor, variableFactors in zip(
                self.columnTitles[:-1],
                self.data[0, :-1],
                self.data[1:, :-1].T.astype(int),
            ):
                if (int(statFactor) != 1) or all(variableFactors == 0):
                    label = f"Global K for {globalK} = {statFactor}"
                    needsCross = True
                else:
                    label = f"Global K for {globalK} ="
                    needsCross = False
                for variable, factor in zip(variables, variableFactors):
                    if factor == 0 or factor == "":
                        continue
                    if needsCross:
                        label += " ×"
                    else:
                        needsCross = True
                    label += f" {variable}"
                    if factor == 1:
                        continue
                    label += str(factor).translate(trans)
                labels.append(label)
            self.equationsLabel.configure(text="\n".join(labels))
        except Exception:
            pass
        return True

    def convertData(self, string):
        return string


class CustomKsPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter relationships between Ks")

        height = int(self.master.winfo_height() * 0.4)
        self.frame = ttk.Frame(self, height=height)
        self.frame.pack(expand=True, fill="both")

        customKsLabel = WrappedLabel(
            self.frame,
            text=(
                "Each row represents a variable that will be optimised. Each column"
                " represents a complex. The global K for each complex is the product of"
                " a statistical factor, and all the variables raised to the exponents"
                " specified in that column.\n\nIn the final column, specify a value to"
                " fix the variable, leave empty to optimise the variable, or write"
                " ~number to provide an initial guess for the optimisation.\n\nThe K"
                " for each complex is the global equilibrium constant. For polymers, K₂"
                " is the constant for the formation of the dimer, and Kₙ the constant"
                " for each subsequent binding."
            ),
            padding=5,
        )
        customKsLabel.pack(expand=False, fill="both")

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

        scrolledFrame = ScrolledFrame(
            self.frame, max_width=self.winfo_toplevel().master.winfo_width() - 200
        )
        scrolledFrame.pack(expand=True, fill="both")
        scrolledFrame.pack(expand=True, fill="both")

        self.innerFrame = scrolledFrame.display_widget(ttk.Frame, stretch=True)

        self.customKsTable = CustomKsTable(self.innerFrame, titration)
        self.customKsTable.pack(expand=True, fill="both")

        self.labelFont = font.nametofont("TkTextFont").copy()
        self.labelFont["size"] = int(1.3 * self.labelFont["size"])

        self.equationsLabel = ttk.Label(
            self.innerFrame, anchor="center", font=self.labelFont, padding=5
        )
        self.equationsLabel.pack(fill="both")
        self.customKsTable.equationsLabel = self.equationsLabel
        self.customKsTable.createLabels()

    def reset(self):
        self.customKsTable.destroy()
        self.customKsTable = CustomKsTable(self.innerFrame, self.titration)
        self.customKsTable.pack(expand=True, fill="both")
        self.customKsTable.label = self.equationsLabel
        self.customKsTable.createLabels()

    def saveData(self):
        self.statisticalFactors = self.customKsTable.data[0, :-1].astype(int)
        self.kNames = self.customKsTable.rowTitles[1:]
        self.ksMatrix = self.customKsTable.data[1:, :-1].astype(int)

        kCells = self.customKsTable.data[1:, -1]
        self.knownKs = ma.array(
            [
                ma.masked if kCell == "" or kCell.startswith("~") else float(kCell)
                for kCell in kCells
            ]
        )
        self.initialKs = ma.array(
            [
                float(kCell[1:]) if kCell.startswith("~") else ma.masked
                for kCell in kCells
            ]
        )

        self.saved = True
        self.destroy()


# TODO: convert to new format
class GetKsCustom(EquilibriumConstants):
    Popup = CustomKsPopup

    popupAttributes = (
        "ksMatrix",
        "statisticalFactors",
        "kNames",
        "knownKs",
        "initialKs",
    )

    def run(self, kVars):
        # print(kVars)
        # microKs as a column vector, with the unknown values filled in
        microKs = self.knownKs.copy()
        microKs[self.knownMask] = kVars
        microKs = np.atleast_2d(microKs).T

        # perform the calculation as previewed in the popup
        globalKs = self.statisticalFactors * np.prod(microKs**self.ksMatrix, 0)
        return globalKs


class KnownKsTable(Table):
    def __init__(self, master, titration):
        self.titration = titration
        super().__init__(
            master,
            0,
            0,
            ["Value"],
            rowOptions=("readonlyTitles",),
            columnOptions=("readonlyTitles",),
            allowGuesses=True,
        )
        self.outputNames = self.titration.speciation.variableNames
        self.populateDefault()

    def populateDefault(self):
        # TODO: knownKs and initialKs should be dicts or structured arrays, in order to
        # still work if the required number of outputs changes
        if len(self.outputNames) != len(self.titration.equilibriumConstants.knownKs):
            for name in self.outputNames:
                self.addRow(name, [""])
            return

        for name, knownK, initialK in zip(
            self.outputNames,
            self.titration.equilibriumConstants.knownKs,
            self.titration.equilibriumConstants.initialKs,
        ):
            if knownK is not ma.masked:
                value = f"{knownK:g}"
            elif initialK is not ma.masked:
                value = f"~{initialK:g}"
            else:
                value = ""
            self.addRow(name, [value])

    def convertData(self, string):
        return string


class KnownKsPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        height = int(self.master.winfo_height() * 0.4)
        frame = ScrolledFrame(
            self,
            height=height,
            max_width=self.winfo_toplevel().master.winfo_width() - 200,
        )
        frame.pack(expand=True, fill="both")

        innerFrame = frame.display_widget(ttk.Frame, stretch=True)
        knownKsLabel = WrappedLabel(
            innerFrame,
            text=(
                "Enter known K values, leave empty to optimise the value, or write"
                " ~number to provide an initial guess for the optimisation.\n\nThe K"
                " for each complex is the global equilibrium constant. For polymers, K₂"
                " is the constant for the formation of the dimer, and Kₙ the constant"
                " for each subsequent binding."
            ),
            padding=5,
        )
        knownKsLabel.pack(expand=False, fill="both")

        self.knownKsTable = KnownKsTable(innerFrame, titration)
        self.knownKsTable.pack(expand=True, fill="both")

        buttonFrame = ButtonFrame(innerFrame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

    def reset(self):
        self.knownKsTable.resetData()
        self.knownKsTable.columnTitles = ["Value"]
        self.knownKsTable.populateDefault()

    def saveData(self):
        kCells = self.knownKsTable.data[:, 0]

        self.knownKs = ma.array(
            [
                ma.masked if kCell == "" or kCell.startswith("~") else float(kCell)
                for kCell in kCells
            ]
        )
        self.initialKs = ma.array(
            [
                float(kCell[1:]) if kCell.startswith("~") else ma.masked
                for kCell in kCells
            ]
        )

        self.saved = True
        self.destroy()


class GetKsKnown(EquilibriumConstants):
    Popup = KnownKsPopup
    popupAttributes = ("knownKs", "initialKs")

    @property
    def kNames(self):
        return self.titration.speciation.variableNames


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Equilibria"
    dropdownLabelText = "Fix any K values?"
    dropdownOptions = {
        "No, optimise all Ks": GetKsAll,
        # "Assume second binding weak": GetKsNonspecific,
        "Fix some known Ks": GetKsKnown,
        "Custom": GetKsCustom,
    }
    attributeName = "equilibriumConstants"



import itertools
import tkinter.ttk as ttk
from abc import abstractmethod

import numpy as np
from numpy import ma
from numpy.linalg import lstsq
from scipy.optimize import lsq_linear

from . import moduleFrame
from .table import ButtonFrame, Table


class FitSignals(moduleFrame.Strategy):
    requiredAttributes = ()

    @abstractmethod
    def leastSquares(self, x, y):
        pass

    def run(self, contributorConcs, knownSpectra):
        hasMissingDatapoints = ma.is_masked(self.titration.processedData)
        hasKnownSpectra = np.any(ma.getmaskarray(knownSpectra) == False)  # noqa: E712
        hasDifferentSignalsPerMolecule = hasattr(
            self.titration.contributingSpecies, "signalToMoleculeMap"
        )

        if hasMissingDatapoints or hasKnownSpectra or hasDifferentSignalsPerMolecule:
            # need to process each signal separately
            explainedData = ma.dot(contributorConcs, knownSpectra).filled(0)
            unexplainedData = self.titration.processedData - explainedData

            fittedSpectra = knownSpectra.copy()

            signalsCount = unexplainedData.shape[1]
            contributorsCount = self.titration.contributors.outputCount
            residuals = np.empty(signalsCount)

            if hasattr(self.titration.contributingSpecies, "signalToMoleculeMap"):
                splitIndices = np.cumsum(
                    self.titration.contributors.contributorsCountPerMolecule
                )[:-1]

                # For each signal, take only the relevant contributors' concentrations
                contributorsSlicePerMolecule = np.split(
                    np.arange(contributorsCount), splitIndices
                )
                contributorsSlicePerSignal = [
                    contributorsSlicePerMolecule[molecule]
                    for molecule in self.titration.contributingSpecies.signalToMoleculeMap
                ]

            else:
                contributorsSlicePerSignal = [
                    np.arange(contributorsCount)
                ] * signalsCount

            for index, (
                signalData,
                signalKnownSpectra,
                signalContributorsSlice,
            ) in enumerate(
                zip(
                    unexplainedData.T,
                    knownSpectra.T,
                    contributorsSlicePerSignal,
                )
            ):
                dataMask = ~ma.getmaskarray(signalData)
                unknownSpectraMask = ma.getmaskarray(
                    signalKnownSpectra[signalContributorsSlice]
                )
                unknownSpectraSlice = signalContributorsSlice[unknownSpectraMask]
                relevantContributorConcs = contributorConcs[dataMask, :][
                    :, unknownSpectraSlice
                ]

                try:
                    (
                        fittedSpectra[unknownSpectraSlice, index],
                        residuals[index],
                    ) = self.leastSquares(
                        relevantContributorConcs, signalData.compressed()
                    )
                except ValueError:
                    # Should only happen with incorrect constraints
                    (fittedSpectra[unknownSpectraSlice, index], _) = self.leastSquares(
                        relevantContributorConcs, signalData.compressed()
                    )
                    residuals[index] = (
                        np.linalg.norm(
                            relevantContributorConcs
                            @ fittedSpectra[unknownSpectraSlice, index]
                            - signalData.compressed(),
                            ord=2,
                        )
                        ** 2
                    )
            fittedCurves = ma.dot(contributorConcs, fittedSpectra)
        else:
            # can process all signals at once
            fittedSpectra, residuals = self.leastSquares(
                contributorConcs, self.titration.processedData
            )
            fittedCurves = ma.dot(contributorConcs, fittedSpectra)

        return fittedSpectra, residuals, fittedCurves


class FitSignalsUnconstrained(FitSignals):
    def leastSquares(self, x, y):
        b, residuals, _, _ = lstsq(x, y, rcond=None)
        return b, residuals


class SignalConstraintsTable(Table):
    def __init__(self, master, titration):
        if hasattr(titration, "signalConstraints"):
            signalConstraints = np.where(
                np.isinf(titration.fitSignals.signalConstraints),
                "",
                titration.fitSignals.signalConstraints,
            )
        else:
            signalConstraints = None

        super().__init__(
            master,
            0,
            0,
            (
                "Lower",
                "Upper",
            ),
            maskBlanks=True,
            rowOptions=(),
            columnOptions="readonlyTitles",
            boldTitles=True,
        )
        self.addRow(data=signalConstraints)


class SignalConstraintsPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter signal constraints")

        self.frame = ttk.Frame(self)
        self.frame.pack(expand=True, fill="both")

        constraintsLabel = ttk.Label(
            self.frame, text="Leave cells blank for no lower/upper bound."
        )
        constraintsLabel.pack()

        self.constraintsTable = SignalConstraintsTable(self.frame, titration)
        self.constraintsTable.pack(expand=True, fill="both")

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

    def reset(self):
        self.constraintsTable.data = np.array([["", ""]])

    def saveData(self):
        constraints = self.constraintsTable.data[0]

        self.signalConstraints = np.where(
            np.isnan(constraints), (-np.inf, np.inf), constraints
        )

        self.saved = True
        self.destroy()


class FitSignalsConstrained(FitSignals):
    requiredAttributes = FitSignals.requiredAttributes + ("signalConstraints",)

    def leastSquares(self, x, y):
        if y.ndim == 1:
            return self.leastSquaresSingle(x, y)
        else:
            b, residuals = zip(*[self.leastSquaresSingle(x, col) for col in y.T])
            return np.array(b).T, np.array(residuals)

    def leastSquaresSingle(self, x, y):
        result = lsq_linear(
            x,
            y,
            self.signalConstraints,
            method="bvls",
        )
        # cost is 0.5 * ||A x - b||**2
        return result.x, 2 * result.cost


class FitSignalsNonnegative(FitSignalsConstrained):
    signalConstraints = np.array([0, np.inf])


class FitSignalsCustom(FitSignalsConstrained):
    Popup = SignalConstraintsPopup
    popupAttributes = ("signalConstraints",)


class FitSignalsODR(FitSignals):
    def run(self, contributorConcs, knownSpectra):
        X = np.asarray(contributorConcs)
        Y = np.asarray(self.titration.processedData)
        relativeWeightYOverX = 100
        scaling = (Y.mean() * Y.size) / (X.mean() * X.size) / relativeWeightYOverX
        m, n = X.shape
        XY = np.hstack([X, Y / scaling])
        U, s, Vh = np.linalg.svd(XY)
        V = Vh.T
        Vxy = V[:n, n:]
        Vyy = V[n:, n:]
        B = scaling * -Vxy @ np.linalg.inv(Vyy)
        EF = -XY @ V[:, n:] @ V[:, n:].T
        E, F = EF[:, :n], EF[:, n:]
        norm = np.linalg.norm(EF, ord="fro")
        fittedCurves = X @ B
        # print(f"{s=}")
        print(f"rank={np.linalg.matrix_rank(Vyy)}")
        print(np.linalg.norm(E), np.linalg.norm(F))
        # breakpoint()
        return B, norm**2, fittedCurves


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Spectra"
    dropdownLabelText = "Apply constraints to fitted spectra?"
    dropdownOptions = {
        "No": FitSignalsUnconstrained,
        "Nonnegative": FitSignalsNonnegative,
        "Custom constraints": FitSignalsCustom,
        "USE ODR": FitSignalsODR,
    }
    attributeName = "fitSignals"



import tkinter as tk
import tkinter.ttk as ttk
from abc import ABC

from . import style
from .style import padding


# all module strategies should be a subclass
class Strategy(ABC):
    Popup = None

    # List of attributes that are set through the popup window, and can be
    # loaded from / saved to a file.
    popupAttributes = ()

    # List of attributes that each base Strategy class should define, for all concrete
    # strategies to set either in __init__ or from the popup. Compliance by the concrete
    # strategies is checked in ModuleFrame.callback().
    requiredAttributes = NotImplemented

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.requiredAttributes is NotImplemented:
            raise NotImplementedError(
                f"Can't define class {cls.__name__} without implementing the abstract"
                " class attribute requiredAttributes"
            )

    def checkAttributes(self):
        for attr in self.requiredAttributes:
            if not hasattr(self, attr):
                raise NotImplementedError(
                    f"Can't set strategy {type(self).__name__} without"
                    f" implementing the required attribute {attr}"
                )

    def __init__(self, titration):
        self.titration = titration

    @property
    def outputCount(self):
        return len(self.outputNames)

    @property
    def variableCount(self):
        return len(self.variableNames)


class Popup(tk.Toplevel):
    def show(self):
        if self._windowingsystem != "aqua":
            # On some versions of macOS, calling grab_set here sets the popup
            # grab before the global grab on the menu is realeased, leaving
            # the popup unresponsive to mouse events.
            # TODO: find better place for callback that avoids this issue
            self.grab_set()
        self.transient(self.master)
        self.saved = False
        self.wait_window()
        return self.saved


class GroupFrame(ttk.LabelFrame):
    def __init__(self, master, group, *args, **kwargs):
        super().__init__(
            master,
            labelanchor="n",
            borderwidth=5,
            *args,
            **kwargs,
        )

        self.labelWidget = ttk.Label(
            self,
            text=group,
            font=style.italicFont,
        )
        self.configure(labelwidget=self.labelWidget)


class ModuleFrame(ttk.Frame):
    group = ""
    dropdownLabelText = ""
    dropdownOptions = {}
    attributeName = ""
    setDefault = True

    def __init__(self, parent, *args, **kwargs):
        try:
            groupFrames = parent.groupFrames
        except AttributeError:
            groupFrames = parent.groupFrames = {}

        try:
            labelFrame = groupFrames[self.group]
        except KeyError:
            labelFrame = groupFrames[self.group] = GroupFrame(parent, self.group)
            labelFrame.grid(sticky="nesw", pady=padding)

        super().__init__(labelFrame, *args, **kwargs)

        self.stringVar = tk.StringVar()

        self.dropdownLabel = ttk.Label(
            self, text=self.dropdownLabelText, justify="left"
        )
        self.dropdownLabel.pack(fill="x")

        strategies = list(self.dropdownOptions.keys())
        self.lastValue = ""
        optionMenu = ttk.OptionMenu(
            self,
            self.stringVar,
            self.lastValue,
            command=self.callback,
            *strategies,
            style="primary.Outline.TMenubutton",
        )
        optionMenu.configure(width=30)

        optionMenu.pack(fill="x", pady=(2, padding))

    def update(self, titration, setDefault=False):
        self.titration = titration
        if setDefault and self.setDefault:
            defaultValue = list(self.dropdownOptions.keys())[0]
            self.stringVar.set(defaultValue)
            self.callback(defaultValue)
            return
        elif setDefault:
            setattr(self.titration, self.attributeName, None)

        if getattr(self.titration, self.attributeName) is None:
            self.stringVar.set("")
            return
        self.lastValue = list(self.dropdownOptions.keys())[
            list(self.dropdownOptions.values()).index(
                type(getattr(self.titration, self.attributeName))
            )
        ]
        self.stringVar.set(self.lastValue)

    def callback(self, value):
        SelectedStrategy = self.dropdownOptions[value]
        selectedStrategy = SelectedStrategy(self.titration)
        if selectedStrategy.Popup is not None:
            root = self.winfo_toplevel()
            popup = selectedStrategy.Popup(self.titration, master=root)
            popup.geometry(f"+{root.winfo_x()+100}+{root.winfo_y()+100}")
            if not popup.show():
                # Returns False if the new options shouldn't be saved, so restore the
                # previous value.
                self.stringVar.set(self.lastValue)
                return

            for attr in selectedStrategy.popupAttributes:
                setattr(selectedStrategy, attr, getattr(popup, attr))

        selectedStrategy.checkAttributes()

        setattr(self.titration, self.attributeName, selectedStrategy)
        self.lastValue = value



import warnings
from pathlib import Path

import setuptools

with open("README.md") as readmeFile:
    long_description = readmeFile.read()


def get_version():
    filePath = Path("./musketeer/__init__.py")
    if not filePath.exists():
        warnings.warn(
            "Could not locate __init__.py to read the version number from",
            RuntimeWarning,
        )
        return "0.0.1.unknown.version"
    with Path("./musketeer/__init__.py").open() as file:
        for line in file:
            if line.startswith("__version__"):
                return line.split('"')[1]
        else:
            warnings.warn(
                "Could not find version string in __init__.py", RuntimeWarning
            )
            return "0.0.1.unknown.version"


setuptools.setup(
    name="musketeer",
    version=get_version(),
    author="Daniil Soloviev",
    author_email="dos23@cam.ac.uk",
    description="A tool for fitting data from titration experiments.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
    url="https://github.com/daniilS/Musketeer",
    packages=["musketeer"],
    include_package_data=True,
    install_requires=[
        "matplotlib >= 3.7.1",
        "numpy",
        "packaging",
        "scipy",
        "ttkbootstrap >=0.5.2, <1.0",
        "ttkwidgets",
        "tksheet",
    ],
    python_requires=">=3.8",
)



from abc import abstractmethod

import numpy as np

from . import moduleFrame


class combineResiduals(moduleFrame.Strategy):
    requiredAttributes = ()

    @abstractmethod
    def run(self, residuals):
        pass


class ResidualsSum(combineResiduals):
    def run(self, residuals):
        return np.sqrt(np.sum(residuals))


class ResidualsSumScaled(combineResiduals):
    # Provided for legacy reasons, not currently recommended
    # sum of the residuals, scaled by the number of data points for each signal
    def run(self, residuals):
        numPoints = self.titration.processedData.count(axis=0)
        return np.sqrt(np.sum(residuals / numPoints))


class ResidualsSumNormalised(combineResiduals):
    def run(self, residuals):
        raise NotImplementedError


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Fitting"
    dropdownLabelText = "Normalise residuals?"
    dropdownOptions = {
        "No": ResidualsSum,
        "By number of points": ResidualsSumScaled,
    }
    attributeName = "combineResiduals"



import csv
import tkinter as tk
import tkinter.filedialog as fd
import tkinter.ttk as ttk
from collections import namedtuple

import numpy as np
from numpy import ma
from tksheet import Sheet

from . import moduleFrame
from .style import padding
from .table import ButtonFrame

Params = namedtuple(
    "Params",
    ("yQuantity", "yUnit", "xQuantity", "xUnit"),
    defaults=[""] * 4,
)

predefinedParams = {
    "UV-Vis": Params("Abs", "AU", "λ", "nm"),
    "NMR": Params("δ", "ppm"),
}


class EditDataPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter data")

        frame = ttk.Frame(self, padding=15)
        frame.pack(expand=True, fill="both")

        paramsFrame = ttk.Frame(frame)
        paramsFrame.pack(expand=False, fill="both")

        paramsFrameLeft = ttk.Frame(paramsFrame)
        paramsFrameLeft.grid(row=0, column=0, sticky="ne", padx=5, pady=5)
        paramsFrame.grid_columnconfigure(0, weight=1)

        self.hasAdditionTitles = tk.BooleanVar()
        self.additionTitlesCheckbutton = ttk.Checkbutton(
            paramsFrameLeft, text="Addition titles", variable=self.hasAdditionTitles
        )
        self.hasSignalTitles = tk.BooleanVar()
        self.signalTitlesCheckbutton = ttk.Checkbutton(
            paramsFrameLeft, text="Signal titles", variable=self.hasSignalTitles
        )

        self.additionsRowsRadiobutton = ttk.Radiobutton(
            paramsFrameLeft, value=0, text="Rows are additions, columns are signals"
        )
        self.additionsColumnsRadiobutton = ttk.Radiobutton(
            paramsFrameLeft, value=1, text="Rows are signals, columns are additions"
        )

        self.additionTitlesCheckbutton.pack(pady=2.5)
        self.signalTitlesCheckbutton.pack(pady=2.5)
        self.additionsRowsRadiobutton.pack(pady=2.5)
        self.additionsColumnsRadiobutton.pack(pady=2.5)

        paramsFrameRight = ttk.Frame(paramsFrame)
        paramsFrameRight.grid(row=0, column=1, sticky="nw", padx=5, pady=5)
        paramsFrame.grid_columnconfigure(1, weight=1)

        self.optionMenuVar = tk.StringVar(self)
        optionMenu = ttk.OptionMenu(
            paramsFrameRight,
            self.optionMenuVar,
            None,
            *predefinedParams.keys(),
            style="Outline.TMenubutton",
            command=self.setParams
        )
        optionMenu.configure(width=max([len(s) for s in predefinedParams]) + 1)
        optionMenu.grid(row=0, columnspan=2, pady=2.5)

        self.yQuantityLabel = ttk.Label(paramsFrameRight, text="Measured quantity:")
        self.yQuantityLabel.grid(row=1, column=0, sticky="w")
        self.yUnitLabel = ttk.Label(paramsFrameRight, text="Unit:")
        self.yUnitLabel.grid(row=1, column=1, sticky="w")
        self.yQuantity = tk.StringVar(self)
        self.yUnit = tk.StringVar(self)
        self.yQuantityWidget = ttk.Entry(paramsFrameRight, textvariable=self.yQuantity)
        self.yQuantityWidget.grid(row=2, column=0, sticky="w")
        self.yUnitWidget = ttk.Entry(
            paramsFrameRight, width=10, textvariable=self.yUnit
        )
        self.yUnitWidget.grid(row=2, column=1, sticky="w")

        self.xQuantityLabel = ttk.Label(
            paramsFrameRight, text="Continuous signals x-axis quantity:"
        )
        self.xQuantityLabel.grid(row=3, column=0, sticky="w")
        self.xUnitLabel = ttk.Label(paramsFrameRight, text="Unit:")
        self.xUnitLabel.grid(row=3, column=1, sticky="w")
        self.xQuantity = tk.StringVar(self)
        self.xUnit = tk.StringVar(self)
        self.xQuantityWidget = ttk.Entry(paramsFrameRight, textvariable=self.xQuantity)
        self.xQuantityWidget.grid(row=4, column=0, sticky="w")
        self.xUnitWidget = ttk.Entry(
            paramsFrameRight, width=10, textvariable=self.xUnit
        )
        self.xUnitWidget.grid(row=4, column=1, sticky="w")

        self.sheet = Sheet(
            frame,
            data=[[]],
            expand_sheet_if_paste_too_big=True,
        )

        self.sheet.pack(side="top", expand=True, fill="both")
        self.sheet.enable_bindings()

        buttonFrame = ButtonFrame(self, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

        loadButton = ttk.Button(buttonFrame, text="Load from CSV", command=self.loadCSV)
        loadButton.pack(side="left", padx=padding)

        self.populateDefault()

    def populateDefault(self):
        titration = self.titration
        self.hasAdditionTitles.set(titration.hasAdditionTitles)
        self.hasSignalTitles.set(titration.hasSignalTitles)

        if not self.titration.transposeData:
            self.additionsRowsRadiobutton.invoke()
        else:
            self.additionsColumnsRadiobutton.invoke()

        self.optionMenuVar.set("")

        for param in Params._fields:
            if hasattr(titration, param):
                getattr(self, param).set(getattr(titration, param))
            else:
                getattr(self, param).set("")

        data = titration.rawData.copy()
        if data.shape == (0, 0):
            data = np.full((16, 5), "")
        else:
            formattedData = data.astype(str)
            formattedData[np.isnan(data)] = ""
            data = formattedData
        if titration.hasAdditionTitles and titration.hasSignalTitles:
            data = np.c_[titration.additionTitles, data]
            data = np.r_[[np.insert(titration.signalTitlesStrings, 0, "")], data]
        elif titration.hasAdditionTitles:
            data = np.c_[titration.additionTitles, data]
        elif titration.hasSignalTitles:
            data = np.r_[[titration.signalTitlesStrings], data]

        if titration.transposeData:
            data = data.T
        self.sheet.set_sheet_data(data.tolist())

    def setParams(self, selection):
        params = predefinedParams[selection]
        for k, v in params._asdict().items():
            if v is not None:
                getattr(self, k).set(v)

    def reset(self):
        self.populateDefault()

    def processData(self):
        titration = self.titration
        data = np.array(self.sheet.get_sheet_data())
        data = data[~np.all(data == "", 1), :][:, ~np.all(data == "", 0)]

        titration.hasSignalTitles = self.signalTitlesCheckbutton.instate(["selected"])
        if not titration.hasSignalTitles:
            titration.continuous = False
        titration.hasAdditionTitles = self.additionTitlesCheckbutton.instate(
            ["selected"]
        )
        titration.transposeData = self.additionsColumnsRadiobutton.instate(["selected"])

        if titration.transposeData:
            data = data.T

        if titration.hasAdditionTitles and titration.hasSignalTitles:
            titration.additionTitles = data[1:, 0]
            titration.signalTitles = data[0, 1:]
            rawData = data[1:, 1:]
        elif titration.hasAdditionTitles:
            titration.additionTitles = data[:, 0]
            rawData = data[:, 1:]
        elif titration.hasSignalTitles:
            titration.signalTitles = data[0, :]
            rawData = data[1:, :]
        else:
            rawData = data

        rawData = rawData.astype(object)  # to allow setting values to "nan"
        rawData[rawData == ""] = "nan"
        return ma.masked_invalid(rawData.astype(float))

    def saveData(self):
        self.titration.rawData = self.processData()
        for param in Params._fields:
            setattr(self.titration, param, getattr(self, param).get())

        self.saved = True
        self.destroy()

    def loadCSV(self):
        fileType = tk.StringVar(self)
        filePath = fd.askopenfilename(
            master=self,
            title="Load from CSV",
            filetypes=[("All files", "*.*"), ("CSV files", "*.csv")],
            typevariable=fileType,
        )
        if filePath == "":
            return
        with open(filePath, encoding="utf-8-sig") as file:
            d = csv.Sniffer().sniff(file.readline() + file.readline())
            file.seek(0)
            data = list(csv.reader(file, dialect=d))
        self.sheet.set_sheet_data(data)
